# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
"""Gitignore parser configured to work for edk2-pytool-library."""

import collections
import os
import re
import sys
from os.path import abspath, dirname
from pathlib import Path
from typing import Callable, Optional, Union

"""Original file is from
https://github.com/mherrmann/gitignore_parser/blob/master/gitignore_parser.py
sha hash: 1b51ef1f058efc8bcdcb063bf7b16d1394f03fc6

Original License:

MIT License

Copyright (c) 2018 Michael Herrmann
Copyright (c) 2015 Steve Cook

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


def handle_negation(file_path: str, rules: list) -> bool:
    """Allows `matched` value override if negation is true.

    Otherwise `matched` cannot be overwritten with an exception.
    Used for ensuring rules with ! will override a previous true result back to false.
    """
    for rule in reversed(rules):
        if rule.match(file_path):
            return not rule.negation
    return False


def parse_gitignore_file(full_path: str, base_dir: Optional[str] = None) -> Callable:
    """Parse a gitignore file."""
    if base_dir is None:
        base_dir = dirname(full_path)
    with open(full_path) as ignore_file:
        lines = ignore_file.readlines()
    return parse_gitignore_lines(lines, full_path, base_dir)


def parse_gitignore_lines(lines: list, full_path: str, base_dir: str) -> Callable:
    """Parse a list of lines matching gitignore syntax."""
    counter = 0
    rules = []
    for line in lines:
        counter += 1
        line = line.rstrip("\n")
        rule = rule_from_pattern(line, base_path=Path(base_dir).resolve(), source=(full_path, counter))
        if rule:
            rules.append(rule)
    if not any(r.negation for r in rules):
        return lambda file_path: any(r.match(file_path) for r in rules)
    else:
        # We have negation rules. We can't use a simple "any" to evaluate them.
        # Later rules override earlier rules.
        return lambda file_path: handle_negation(file_path, rules)


def rule_from_pattern(pattern: str, base_path: Optional[str] = None, source: Optional[str] = None) -> "IgnoreRule":
    """Generates an IgnoreRule object from a pattern.

    Take a .gitignore match pattern, such as "*.py[cod]" or "**/*.bak",
    and return an IgnoreRule suitable for matching against files and
    directories. Patterns which do not match files, such as comments
    and blank lines, will return None.
    Because git allows for nested .gitignore files, a base_path value
    is required for correct behavior. The base path should be absolute.
    """
    if base_path and base_path != Path(base_path).resolve():
        raise ValueError("base_path must be absolute")
    # Store the exact pattern for our repr and string functions
    orig_pattern = pattern
    # Early returns follow
    # Discard comments and separators
    if pattern.strip() == "" or pattern[0] == "#":
        return
    # Strip leading bang before examining double asterisks
    if pattern[0] == "!":
        negation = True
        pattern = pattern[1:]
    else:
        negation = False
    # Multi-asterisks not surrounded by slashes (or at the start/end) should
    # be treated like single-asterisks.
    pattern = re.sub(r"([^/])\*{2,}", r"\1*", pattern)
    pattern = re.sub(r"\*{2,}([^/])", r"*\1", pattern)

    # Special-casing '/', which doesn't match any files or directories
    if pattern.rstrip() == "/":
        return

    directory_only = pattern[-1] == "/"
    # A slash is a sign that we're tied to the base_path of our rule
    # set.
    anchored = "/" in pattern[:-1]
    if pattern[0] == "/":
        pattern = pattern[1:]
    if pattern[0] == "*" and len(pattern) >= 2 and pattern[1] == "*":
        pattern = pattern[2:]
        anchored = False
    if pattern[0] == "/":
        pattern = pattern[1:]
    if pattern[-1] == "/":
        pattern = pattern[:-1]
    # patterns with leading hashes or exclamation marks are escaped with a
    # backslash in front, unescape it
    if pattern[0] == "\\" and pattern[1] in ("#", "!"):
        pattern = pattern[1:]
    # trailing spaces are ignored unless they are escaped with a backslash
    i = len(pattern) - 1
    striptrailingspaces = True
    while i > 1 and pattern[i] == " ":
        if pattern[i - 1] == "\\":
            pattern = pattern[: i - 1] + pattern[i:]
            i = i - 1
            striptrailingspaces = False
        else:
            if striptrailingspaces:
                pattern = pattern[:i]
        i = i - 1
    regex = fnmatch_pathname_to_regex(pattern, directory_only, negation, anchored=bool(anchored))
    return IgnoreRule(
        pattern=orig_pattern,
        regex=regex,
        negation=negation,
        directory_only=directory_only,
        anchored=anchored,
        base_path=_normalize_path(base_path) if base_path else None,
        source=source,
    )


IGNORE_RULE_FIELDS = [
    "pattern",
    "regex",  # Basic values
    "negation",
    "directory_only",
    "anchored",  # Behavior flags
    "base_path",  # Meaningful for gitignore-style behavior
    "source",  # (file, line) tuple for reporting
]


class IgnoreRule(collections.namedtuple("IgnoreRule_", IGNORE_RULE_FIELDS)):
    """Class representing a single rule parsed from a .ignore file."""

    def __str__(self) -> str:
        """String representation (user friendly) of the rule."""
        return self.pattern

    def __repr__(self) -> str:
        """String representation (developer friendly) of the rule."""
        return "".join(["IgnoreRule('", self.pattern, "')"])

    def match(self, abs_path: str) -> bool:
        """Returns True or False if the path matches the rule."""
        matched = False
        if self.base_path:
            try:
                rel_path = _normalize_path(abs_path).relative_to(self.base_path).as_posix()
            except ValueError as e:
                if "is not in the subpath of" in str(e):
                    return False
                else:
                    raise
        else:
            rel_path = _normalize_path(abs_path).as_posix()
        # Path() strips the trailing following symbols on windows, so we need to
        # preserve it: ' ', '.'
        if sys.platform.startswith("win"):
            rel_path += " " * _count_trailing_symbol(" ", abs_path)
            rel_path += "." * _count_trailing_symbol(".", abs_path)
        # Path() strips the trailing slash, so we need to preserve it
        # in case of directory-only negation
        if self.negation and isinstance(abs_path, str) and abs_path[-1] == "/":
            rel_path += "/"
        if rel_path.startswith("./"):
            rel_path = rel_path[2:]
        if re.search(self.regex, rel_path):
            matched = True
        return matched


# Frustratingly, python's fnmatch doesn't provide the FNM_PATHNAME
# option that .gitignore's behavior depends on.
def fnmatch_pathname_to_regex(pattern: str, directory_only: bool, negation: bool, anchored: bool = False) -> str:
    """Implements fnmatch style-behavior, as though with FNM_PATHNAME flagged.

    WARNING: the path seperator will not match shell-style '*' and '.' wildcards.
    """
    i, n = 0, len(pattern)

    seps = [re.escape(os.sep)]
    if os.altsep is not None:
        seps.append(re.escape(os.altsep))
    seps_group = "[" + "|".join(seps) + "]"
    nonsep = r"[^{}]".format("|".join(seps))

    res = []
    while i < n:
        c = pattern[i]
        i += 1
        if c == "*":
            try:
                if pattern[i] == "*":
                    i += 1
                    if i < n and pattern[i] == "/":
                        i += 1
                        res.append("".join(["(.*", seps_group, ")?"]))
                    else:
                        res.append(".*")
                else:
                    res.append("".join([nonsep, "*"]))
            except IndexError:
                res.append("".join([nonsep, "*"]))
        elif c == "?":
            res.append(nonsep)
        elif c == "/":
            res.append(seps_group)
        elif c == "[":
            j = i
            if j < n and pattern[j] == "!":
                j += 1
            if j < n and pattern[j] == "]":
                j += 1
            while j < n and pattern[j] != "]":
                j += 1
            if j >= n:
                res.append("\\[")
            else:
                stuff = pattern[i:j].replace("\\", "\\\\").replace("/", "")
                i = j + 1
                if stuff[0] == "!":
                    stuff = "".join(["^", stuff[1:]])
                elif stuff[0] == "^":
                    stuff = "".join("\\" + stuff)
                res.append("[{}]".format(stuff))
        else:
            res.append(re.escape(c))
    if anchored:
        res.insert(0, "^")
    else:
        res.insert(0, f"(^|{seps_group})")
    if not directory_only:
        res.append("$")
    elif directory_only and negation:
        res.append("/$")
    else:
        res.append("($|\\/)")
    return "".join(res)


def _normalize_path(path: Union[str, Path]) -> Path:
    """Normalize a path without resolving symlinks.

    This is equivalent to `Path.resolve()` except that it does not resolve symlinks.
    Note that this simplifies paths by removing double slashes, `..`, `.` etc. like
    `Path.resolve()` does.
    """
    return Path(abspath(path))


def _count_trailing_symbol(symbol: str, text: str) -> int:
    """Count the number of trailing characters in a string."""
    count = 0
    for char in reversed(str(text)):
        if char == symbol:
            count += 1
        else:
            break
    return count



# @file targettxt_parser.py
# Code to help parse Edk2 Conf/Target.txt file
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code to help parse Edk2 Conf/Target.txt file."""

import os

from edk2toollib.uefi.edk2.parsers.base_parser import HashFileParser


class TargetTxtParser(HashFileParser):
    """Parser for the Edk2 Conf/Target.txt file.

    Attributes:
        Parsed (bool): Whether the object has parsed a file or not
        Lines (list): Ordered list of each line in the file
        Dict (dict): Key / Value pair of all lines that contain a `=` in them (key=value)
        Path (str): path to Target.txt file
    """

    def __init__(self) -> "TargetTxtParser":
        """Inits an empty parser."""
        HashFileParser.__init__(self, "TargetTxtParser")
        self.Lines = []
        self.Parsed = False
        self.Dict = {}
        self.Path = ""

    def ParseFile(self, filepath: str) -> None:
        """Parses the file provided."""
        self.Logger.debug("Parsing file: %s" % filepath)
        if not os.path.isabs(filepath):
            fp = self.FindPath(filepath)
        else:
            fp = filepath
        self.Path = fp
        f = open(fp, "r")
        self.Lines = f.readlines()
        f.close()

        for line in self.Lines:
            sline = self.StripComment(line)

            if sline is None or len(sline) < 1:
                continue

            if sline.count("=") == 1:
                tokens = sline.split("=", 1)
                self.Dict[tokens[0].strip()] = tokens[1].strip()
                self.Logger.debug("Key,values found:  %s = %s" % (tokens[0].strip(), tokens[1].strip()))
                continue

        self.Parsed = True



# @file edk2_db.py
# A class for interacting with a database implemented using json.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A class for interacting with a database implemented using json."""

import logging
import time
import uuid
from contextlib import contextmanager
from typing import Any

from sqlalchemy import create_engine
from sqlalchemy.orm import DeclarativeBase, Session

from edk2toollib.uefi.edk2.path_utilities import Edk2Path


class Base(DeclarativeBase):
    """The base class for creating database table models.

    This class should be the subclass for any table model that will be used with Edk2DB.
    """


class Edk2DB:
    """A SQLite3 database manager for a EDKII workspace.

    This class provides the ability to register parsers that will create / update tables in the database. This will
    create a SQLite datbase file that can be queried using any SQLite3 client. VSCode provides multiple extensions
    for viewing and interacting with the database. Queries can also be created and run in python using the sqlite3
    module that comes with python.

    Edk2DB can, and should, be used as a context manager to ensure that the database is closed properly. If
    not using as a context manager, the `db.connection.commit()` and `db.connection.close()` must be used to cleanly
    close the database.

    Attributes:
        connection (sqlite3.Connection): The connection to the database

    !!! note
        Edk2DB provides a table called `junction` that can be used to make associations between tables. It has the
        following schema: `env_id, table1, key1, table2, key2`.

    Example:
        ```python
        from edk2toollib.database.parsers import *
        table = "..."
        with Edk2DB(Path("path/to/db.db"), edk2path) as db:
            db.register(Parser1(), Parser2(), Parser3())
            db.parse()
            db.connection.execute("SELECT * FROM ?", table)
        ```
    """

    Base = Base

    def __init__(self: "Edk2DB", db_path: str, pathobj: Edk2Path = None, **kwargs: dict[str, Any]) -> "Edk2DB":
        """Initializes the database.

        Args:
            db_path: Path to create or load the database from
            pathobj: Edk2Path object for the workspace
            **kwargs: None
        """
        self.pathobj = pathobj
        self.clear_parsers()
        self.engine = create_engine(f"sqlite:///{db_path}", **kwargs)
        self.Base.metadata.create_all(self.engine)

    @contextmanager
    def session(self) -> Session:
        """Provides a context manager for a session with the database.

        Handles commiting changes and rolling back if an exception is raised.
        """
        session = Session(self.engine)
        try:
            yield session
            session.commit()
        except Exception as e:
            session.rollback()
            raise e
        finally:
            session.close()

    def register(self, *parsers: "TableGenerator") -> None:
        """Registers a one or more table generators.

        Args:
            *parsers: One or more instantiated TableGenerator object
        """
        for parser in parsers:
            self._parsers.append(parser)

    def clear_parsers(self) -> None:
        """Empties the list of registered table generators."""
        self._parsers = []

    def parse(self, env: dict) -> None:
        """Runs all registered table parsers against the database.

        !!! note
            To enable queries to differentiate between two parses, an environment table is always created if it does
            not exist, and a row is added for each call of this command.
        """
        id = str(uuid.uuid4().hex)

        # Fill all tables
        for table in self._parsers:
            logging.debug(f"[{table.__class__.__name__}] starting...")
            t = time.time()
            with self.session() as session:
                table.parse(session, self.pathobj, id, env)
            logging.debug(f"Finished in {round(time.time() - t, 2)}")


class TableGenerator:
    """An interface for a parser that generates a sqlite3 table maintained by Edk2DB.

    Allows you to parse a workspace, file, etc, and load the contents into the database as rows in a table.

    Edk2Db provides a connection to a sqlite3 database and will commit any changes made during `parse` once
    the parser has finished executing and has returned. Review sqlite3 documentation for more information on
    how to interact with the database.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> "TableGenerator":
        """Initialize the query with the specific settings."""

    def parse(self, session: Session, pathobj: Edk2Path, id: str, env: dict) -> None:
        """Execute the parser and update the database."""
        raise NotImplementedError



# @file inf_parser.py
# Code to help parse EDK2 INF files
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code to help parse EDK2 INF files."""

import os
import re

from edk2toollib.uefi.edk2.parsers.base_parser import HashFileParser

AllPhases = [
    "SEC",
    "PEIM",
    "PEI_CORE",
    "DXE_DRIVER",
    "DXE_CORE",
    "DXE_RUNTIME_DRIVER",
    "UEFI_DRIVER",
    "SMM_CORE",
    "DXE_SMM_DRIVER",
    "UEFI_APPLICATION",
    "MM_STANDALONE",
    "MM_CORE_STANDALONE",
    "HOST_APPLICATION",
]


class InfParser(HashFileParser):
    """Object representing a parsed INF with capabilities to parse an INF.

    Attributes:
        Parsed (bool): Whether the object contains a parsed INF
        Lines (list[str]): ordered list of all lines in the INF
        Dict (dict): Key / Value pairs found in the INF.
        LibraryClass (str): library class of the INF
        SupportedPhases (list[str]): list of supported phases (i.e. "SEC", "PEIM", etc.)
        PackagesUsed (list[str]): list of packages used
        LibrariesUsed (list[str]): list of libraries used
        ProtocolsUsed (list[str]): list of protocols used
        GuidsUsed (list[str]): list of guids used
        PpisUsed (list[str]): list of Ppis used
        PcdsUsed (list[str]): list of Pcds used
        Sources (list[str]): list of source files used
        Binaries (list[str]): list of binaries used
        Path (str): Path to the INF file

    NOTE: Key / Value pairs determined by lines that contain a single =
    """

    SECTION_REGEX = re.compile(r"\[(.*)\]")
    SECTION_LIBRARY = re.compile(r"libraryclasses(?:\.([^,.\]]+))?[,.\]]", re.IGNORECASE)
    SECTION_SOURCE = re.compile(r"sources(?:\.([^,.\]]+))?[,.\]]", re.IGNORECASE)

    def __init__(self) -> "InfParser":
        """Inits an empty parser."""
        HashFileParser.__init__(self, "ModuleInfParser")
        self.Lines = []
        self.Parsed = False
        self.Dict = {}
        self.LibraryClass = ""
        self.SupportedPhases = []
        self.PackagesUsed = []
        self.LibrariesUsed = []
        self.ScopedLibraryDict = {}
        self.ScopedSourceDict = {}
        self.ProtocolsUsed = []
        self.GuidsUsed = []
        self.PpisUsed = []
        self.PcdsUsed = []
        self.Sources = []
        self.Binaries = []
        self.Path = ""

    def get_libraries(self, arch_list: list[str]) -> list[str]:
        """Returns a list of libraries depending on the requested archs."""
        libraries = self.ScopedLibraryDict.get("common", []).copy()

        for arch in arch_list:
            libraries = libraries + self.ScopedLibraryDict.get(arch.lower(), []).copy()
        return list(set(libraries))

    def get_sources(self, arch_list: list[str]) -> list[str]:
        """Returns a list of sources depending on the requested archs."""
        sources = self.ScopedSourceDict.get("common", []).copy()

        for arch in arch_list:
            sources = sources + self.ScopedSourceDict.get(arch.lower(), []).copy()
        return list(set(sources))

    def ParseFile(self, filepath: str) -> None:
        """Parses the INF file provided."""
        self.Logger.debug("Parsing file: %s" % filepath)
        if not os.path.isabs(filepath):
            fp = self.FindPath(filepath)
        else:
            fp = filepath
        self.Path = fp
        f = open(fp, "r")
        self.Lines = f.readlines()
        f.close()
        InDefinesSection = False
        InPackagesSection = False
        InLibraryClassSection = False
        InProtocolsSection = False
        InGuidsSection = False
        InPpiSection = False
        InPcdSection = False
        InSourcesSection = False
        InBinariesSection = False

        for line in self.Lines:
            sline = self.StripComment(line)

            if sline is None or len(sline) < 1:
                continue

            sline = self.ReplaceVariables(sline)

            if sline.strip().startswith("DEFINE"):
                tokens = sline.strip().replace("DEFINE", "").split("=", 1)
                self.LocalVars[tokens[0].strip()] = tokens[1].strip()
                self.Logger.info(f"Key,values found for local vars: {tokens[0].strip()}, {tokens[1].strip()}")
                continue

            elif InDefinesSection:
                if sline.strip()[0] == "[":
                    InDefinesSection = False
                else:
                    if sline.count("=") == 1:
                        tokens = sline.split("=", 1)
                        self.Dict[tokens[0].strip()] = tokens[1].strip()
                        #
                        # Parse Library class and phases in special manor
                        #
                        if tokens[0].strip().lower() == "library_class":
                            self.LibraryClass = tokens[1].partition("|")[0].strip()
                            self.Logger.debug("Library class found")
                            if len(tokens[1].partition("|")[2].strip()) < 1:
                                self.SupportedPhases = AllPhases
                            elif tokens[1].partition("|")[2].strip().lower() == "base":
                                self.SupportedPhases = AllPhases
                            else:
                                self.SupportedPhases = tokens[1].partition("|")[2].strip().split()

                        self.Logger.debug("Key,values found:  %s = %s" % (tokens[0].strip(), tokens[1].strip()))

                        continue

            elif InPackagesSection:
                if sline.strip()[0] == "[":
                    InPackagesSection = False
                else:
                    self.PackagesUsed.append(sline.partition("|")[0].strip())
                    continue

            elif InLibraryClassSection:
                if sline.strip()[0] == "[":
                    InLibraryClassSection = False
                else:
                    self.LibrariesUsed.append(sline.partition("|")[0].strip())
                    continue

            elif InProtocolsSection:
                if sline.strip()[0] == "[":
                    InProtocolsSection = False
                else:
                    self.ProtocolsUsed.append(sline.partition("|")[0].strip())
                    continue

            elif InGuidsSection:
                if sline.strip()[0] == "[":
                    InGuidsSection = False
                else:
                    self.GuidsUsed.append(sline.partition("|")[0].strip())
                    continue

            elif InPcdSection:
                if sline.strip()[0] == "[":
                    InPcdSection = False
                else:
                    self.PcdsUsed.append(sline.partition("|")[0].strip())
                    continue

            elif InPpiSection:
                if sline.strip()[0] == "[":
                    InPpiSection = False
                else:
                    self.PpisUsed.append(sline.partition("|")[0].strip())
                    continue

            elif InSourcesSection:
                if sline.strip()[0] == "[":
                    InSourcesSection = False
                else:
                    self.Sources.append(sline.partition("|")[0].strip())
                    continue

            elif InBinariesSection:
                if sline.strip()[0] == "[":
                    InBinariesSection = False
                else:
                    self.Binaries.append(sline.partition("|")[0].strip())
                    continue

            # check for different sections
            if sline.strip().lower().startswith("[defines"):
                InDefinesSection = True

            elif sline.strip().lower().startswith("[packages"):
                InPackagesSection = True

            elif sline.strip().lower().startswith("[libraryclasses"):
                InLibraryClassSection = True

            elif sline.strip().lower().startswith("[protocols"):
                InProtocolsSection = True

            elif sline.strip().lower().startswith("[ppis"):
                InPpiSection = True

            elif sline.strip().lower().startswith("[guids"):
                InGuidsSection = True

            elif (
                sline.strip().lower().startswith("[pcd")
                or sline.strip().lower().startswith("[patchpcd")
                or sline.strip().lower().startswith("[fixedpcd")
                or sline.strip().lower().startswith("[featurepcd")
            ):
                InPcdSection = True

            elif sline.strip().lower().startswith("[sources"):
                InSourcesSection = True

            elif sline.strip().lower().startswith("[binaries"):
                InBinariesSection = True

        # Re-parse for scoped information
        # Hopefully eventually replace all of the above with this
        # logic
        current_section = ""
        arch_list = []

        for line in self.Lines:
            line = self.StripComment(line)

            if line.strip() == "":
                continue

            if line.strip().startswith("#"):
                continue

            if line.strip().startswith("DEFINE"):
                tokens = line.strip().replace("DEFINE", "").split("=", 1)
                self.LocalVars[tokens[0].strip()] = tokens[1].strip()
                self.Logger.info(f"Key,values found for local vars: {tokens[0].strip()}, {tokens[1].strip()}")
                continue

            line = self.ReplaceVariables(line)

            match = self.SECTION_LIBRARY.findall(line)
            if match:
                current_section = "LibraryClasses"
                arch_list = match
                continue
            match = self.SECTION_SOURCE.findall(line)
            if match:
                current_section = "Sources"
                arch_list = match
                continue

            # Catch all others
            match = self.SECTION_REGEX.match(line)
            if match:
                current_section = match.group(1)
                arch_list = []

            # Handle lines when we are in a library section
            if current_section == "LibraryClasses":
                for arch in arch_list:
                    arch = "common" if arch == "" else arch.lower()
                    if arch in self.ScopedLibraryDict:
                        self.ScopedLibraryDict[arch].append(line.split()[0].strip())
                    else:
                        self.ScopedLibraryDict[arch] = [line.split()[0].strip()]

            # Handle lines when we are in a source section
            if current_section == "Sources":
                for arch in arch_list:
                    arch = "common" if arch == "" else arch.lower()
                    src = line.split()[0].strip().rstrip("|")
                    if arch in self.ScopedSourceDict:
                        self.ScopedSourceDict[arch].append(src)
                    else:
                        self.ScopedSourceDict[arch] = [src]
        self.Parsed = True



##
# Handle basic logging outputting to files
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module for handling basically logging to files."""

import logging


class FileHandler(logging.FileHandler):
    """Object for handling basic logging output to files."""

    def __init__(self, filename: str, mode: str = "w+", encoding="utf-8") -> "FileHandler":
        """Init a file handler for the specified file."""
        logging.FileHandler.__init__(self, filename, mode=mode, encoding=encoding)

    def handle(self, record: logging.LogRecord) -> bool:
        """Conditionally emit the specified logging record.

        Emission depends on filters which may have been added to the handler.
        Wrap the actual emission of the record with acquisition/release of
        the I/O thread lock. Returns whether the filter passed the record for
        emission.
        """
        rv = self.filter(record)
        if rv and record.levelno >= self.level:
            self.acquire()
            try:
                self.emit(record)
            finally:
                self.release()
        return rv



# @file dsc_translator
# Translates a DSC object into a file
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
"""Translates a DSC object into a file."""

import logging
import os

from edk2toollib.uefi.edk2.build_objects.dsc import (
    build_option,
    component,
    definition,
    dsc,
    dsc_set,
    library_class,
    pcd,
    pcd_typed,
    pcd_variable,
    sku_id,
)


class DscTranslator:
    """A class used to translate DSCs."""

    @classmethod
    def dsc_to_file(cls: "DscTranslator", dsc_obj: dsc, filepath: str) -> None:
        """Transforms the DSC object to a file."""
        file_path = os.path.abspath(filepath)
        f = open(file_path, "w")
        lines = cls._GetDscLinesFromDscObj(dsc_obj)
        for line in lines:
            f.write(line + "\n")
        f.close()

    @classmethod
    def _GetDscLinesFromDscObj(cls: "DscTranslator", obj: dsc, depth: int = 0) -> list:
        """Gets the DSC strings for an data model objects."""
        lines = []
        depth_pad = "".ljust(depth)
        org_depth = depth
        depth += 2

        if isinstance(obj, (list, set, dsc_set)):
            for item in obj:
                lines += cls._GetDscLinesFromDscObj(item, org_depth)
        elif type(obj) is dsc:
            lines.append(f"{depth_pad}[Defines]")
            lines += cls._GetDscLinesFromDscObj(obj.defines, depth)

            # Second do the Skus
            lines.append(f"{depth_pad}[SkuIds]")
            for x in obj.skus:
                lines += cls._GetDscLinesFromDscObj(x, depth)

            # Third, library classes
            for header, x in obj.library_classes.items():
                lines.append(f"{depth_pad}[LibraryClasses{header}]")
                lines += cls._GetDscLinesFromDscObj(x, depth)

            # Next do the components
            for header, x in obj.components.items():
                lines.append(f"{depth_pad}[Components{header}]")
                lines += cls._GetDscLinesFromDscObj(x, depth)

            # Then PCD's
            for header, x in obj.pcds.items():
                lines.append(f"{depth_pad}[{header}]")
                lines += cls._GetDscLinesFromDscObj(x, depth)

            # Then Build Options
            print(obj.build_options.items())
            for header, x in obj.build_options.items():
                lines.append(f"{depth_pad}[BuildOptions{header}]")
                lines += cls._GetDscLinesFromDscObj(x, depth)

        elif type(obj) is sku_id:
            lines.append(f"{depth_pad}{obj.id}|{obj.name}|{obj.parent}")
        elif type(obj) is library_class:
            lines.append(f"{depth_pad}{obj.libraryclass}|{obj.inf}")
        elif type(obj) is definition:
            def_str = f"{obj.name} =\t{obj.value}"
            if obj.local:
                def_str = "DEFINE " + def_str
            lines.append(depth_pad + def_str)

        elif type(obj) is component:
            lines += cls._FormatComponent(obj, depth)

        elif type(obj) is pcd:
            lines.append(f"{depth_pad}{obj.namespace}.{obj.name}|{obj.value}")

        elif type(obj) is pcd_typed:
            pcd_str = f"{depth_pad}{obj.namespace}.{obj.name}|{obj.value}|{obj.datum_type}"
            if obj.max_size > 0:
                pcd_str += f"|{obj.max_size}"
            lines.append(pcd_str)

        elif type(obj) is pcd_variable:
            pcd_name = f"{depth_pad}{obj.namespace}.{obj.name}|{obj.var_name}"
            if obj.default is None:
                lines.append(f"{pcd_name}|{obj.var_guid}|{obj.var_offset}")
            elif len(obj.attributes) == 0:
                lines.append(f"{pcd_name}|{obj.var_guid}|{obj.var_offset}|{obj.default}")
            else:
                attr = ", ".join(obj.attributes)
                lines.append(f"{pcd_name}|{obj.var_guid}|{obj.var_offset}|{obj.default}|{attr}")

        elif type(obj) is build_option:
            rep = depth_pad if obj.family is None else f"{depth_pad}{obj.family}:"
            rep += "_".join((obj.target, obj.tagname, obj.arch, obj.tool_code, obj.attribute))
            rep += f"= {obj.data}"
            lines.append(rep)
        else:
            logging.warning(f"UNKNOWN OBJECT {obj}")
        return lines

    @classmethod
    def _FormatComponent(cls: "DscTranslator", comp: dsc, depth: int = 0) -> list[str]:
        has_subsection = (
            len(comp.pcds) > 0 or len(comp.defines) > 0 or len(comp.build_options) > 0 or len(comp.library_classes) > 0
        )
        depth_pad = "".ljust(depth)
        if not has_subsection:
            return [
                f"{depth_pad}{comp.inf}",
            ]
        lines = []
        org_depth_pad = depth_pad
        depth_pad += "   "  # add two more onto our pad
        depth += 4
        lines.append(f"{org_depth_pad}{comp.inf} {{")
        if len(comp.pcds) > 0:
            for section, pcds in comp.pcds.items():
                lines.append(f"{depth_pad}<{section}>")
                lines += cls._GetDscLinesFromDscObj(pcds, depth)
        if len(comp.library_classes) > 0:
            lines.append(f"{depth_pad}<LibraryClasses>")
            lines += cls._GetDscLinesFromDscObj(comp.library_classes, depth)
        if len(comp.defines) > 0:
            lines.append(f"{depth_pad}<Defines>")
            lines += cls._GetDscLinesFromDscObj(comp.defines, depth)
        if len(comp.build_options) > 0:
            lines.append(f"{depth_pad}<BuildOptions>")
            lines += cls._GetDscLinesFromDscObj(comp.build_options, depth)
        lines.append(f"{org_depth_pad}}}")
        return lines



# @file
# Module contains helper classes for working with Fault Tolerant Working block content
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module contains helper classes for working with FaultTolerant Working block content."""

import struct
import sys
import uuid
from typing import IO

#
# UEFI GUIDs
#
EdkiiWorkingBlockSignatureGuid = uuid.UUID(fields=(0x9E58292B, 0x7C68, 0x497D, 0xA0, 0xCE, 0x6500FD9F1B95))


class EfiFtwWorkingBlockHeader(object):
    """Object representing a EFI_FAULT_TOLERANT_WORKING_BLOCK_HEADER struct.

    The EDKII Fault tolerant working block header.
    The header is immediately followed by the write queue data.

    typedef struct {
        EFI_GUID  Signature;
        UINT32    Crc;
        UINT8     WorkingBlockValid : 1;
        UINT8     WorkingBlockInvalid : 1;
        UINT8     Reserved : 6;
        UINT8     Reserved3[3];
        UINT64    WriteQueueSize;
    } EFI_FAULT_TOLERANT_WORKING_BLOCK_HEADER;
    """

    def __init__(self) -> "EfiFtwWorkingBlockHeader":
        """Initializes an empty object."""
        self.StructString = "=16sLBBBBQ"  # spell-checker: disable-line
        self.Signature = None
        self.Crc = None
        self.WorkingBlockValidFields = None
        self.Reserved1 = None
        self.Reserved2 = None
        self.Reserved3 = None
        self.WriteQueueSize = None

    def load_from_file(self, file: IO) -> "EfiFtwWorkingBlockHeader":
        """Loads an EFI_FAULT_TOLERANT_WORKING_BLOCK_HEADER from a file.

        Args:
            file (obj): An already opened file.

        Returns:
            (EfiFtwWorkingBlockHeader): self

        Raises:
            (Exception): unknown signature
        """
        # This function assumes that the file has been seeked
        # to the correct starting location.
        orig_seek = file.tell()
        struct_bytes = file.read(struct.calcsize(self.StructString))
        file.seek(orig_seek)

        # Load this object with the contents of the data.
        (
            signature_bin,
            self.Crc,
            self.WorkingBlockValidFields,
            self.Reserved1,
            self.Reserved2,
            self.Reserved3,
            self.WriteQueueSize,
        ) = struct.unpack(self.StructString, struct_bytes)

        # Update the GUID to be a UUID object.
        if sys.byteorder == "big":
            self.Signature = uuid.UUID(bytes=signature_bin)
        else:
            self.Signature = uuid.UUID(bytes_le=signature_bin)

        # Check that signature is valid
        if self.Signature != EdkiiWorkingBlockSignatureGuid:
            raise Exception("FTW Working Block Header has unknown signature: %s" % self.Signature)

        return self

    def serialize(self) -> bytes:
        r"""Serializes the EFI_FAULT_TOLERANT_WORKING_BLOCK_HEADER.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        signature_bin = self.Signature.bytes if sys.byteorder == "big" else self.Signature.bytes_le
        return struct.pack(
            self.StructString,
            signature_bin,
            self.Crc,
            self.WorkingBlockValidFields,
            self.Reserved1,
            self.Reserved2,
            self.Reserved3,
            self.WriteQueueSize,
        )


class EfiFtwWriteHeader(object):
    """Object representing a EFI_FAULT_TOLERANT_WRITE_HEADER struct.

    EFI Fault tolerant block update write queue entry.

    typedef struct {
        UINT8     HeaderAllocated : 1;
        UINT8     WritesAllocated : 1;
        UINT8     Complete : 1;
        UINT8     Reserved : 5;
        EFI_GUID  CallerId;
        UINT64    NumberOfWrites;
        UINT64    PrivateDataSize;
    } EFI_FAULT_TOLERANT_WRITE_HEADER;
    """

    def __init__(self) -> "EfiFtwWriteHeader":
        """Initializes an empty object."""
        self.StructString = "=BBBB16sLQQ"
        self.StatusBits = None
        self.ReservedByte1 = None
        self.ReservedByte2 = None
        self.ReservedByte3 = None
        self.CallerId = None
        self.ReservedUint32 = None
        self.NumberOfWrites = None
        self.PrivateDataSize = None

    def load_from_file(self, file: IO) -> "EfiFtwWriteHeader":
        """Loads an EFI_FAULT_TOLERANT_WRITE_HEADER from a file.

        Args:
            file (obj): An already opened file.

        Returns:
            (EfiFtwWriteHeader): self
        """
        # This function assumes that the file has been seeked
        # to the correct starting location.
        orig_seek = file.tell()
        struct_bytes = file.read(struct.calcsize(self.StructString))
        file.seek(orig_seek)

        # Load this object with the contents of the data.
        (
            self.StatusBits,
            self.ReservedByte1,
            self.ReservedByte2,
            self.ReservedByte3,
            self.CallerId,
            self.ReservedUint32,
            self.NumberOfWrites,
            self.PrivateDataSize,
        ) = struct.unpack(self.StructString, struct_bytes)

        return self

    def serialize(self) -> bytes:
        r"""Serializes the EFI_FAULT_TOLERANT_WRITE_HEADER.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        return struct.pack(
            self.StructString,
            self.StatusBits,
            self.ReservedByte1,
            self.ReservedByte2,
            self.ReservedByte3,
            self.CallerId,
            self.ReservedUint32,
            self.NumberOfWrites,
            self.PrivateDataSize,
        )


class EfiFtwWriteRecord(object):
    """Object representing a EFI_FAULT_TOLERANT_WRITE_RECORD struct.

    EFI Fault tolerant block update write queue record.

    typedef struct {
        UINT8   BootBlockUpdate : 1;
        UINT8   SpareComplete : 1;
        UINT8   DestinationComplete : 1;
        UINT8   Reserved : 5;
        EFI_LBA Lba;
        UINT64  Offset;
        UINT64  Length;
        INT64   RelativeOffset;
    } EFI_FAULT_TOLERANT_WRITE_RECORD;
    """

    def __init__(self) -> "EfiFtwWriteRecord":
        """Initializes an empty object."""
        self.StructString = "=BBBBLQQQQ"  # spell-checker: disable-line
        self.StatusBits = None
        self.ReservedByte1 = None
        self.ReservedByte2 = None
        self.ReservedByte3 = None
        self.ReservedUint32 = None
        self.Lba = None
        self.Offset = None
        self.Length = None
        self.RelativeOffset = None

    def load_from_file(self, file: IO) -> "EfiFtwWriteRecord":
        """Loads an EFI_FAULT_TOLERANT_WRITE_RECORD from a file.

        Args:
            file (obj): An already opened file.

        Returns:
            (EfiFtwWriteRecord): self
        """
        # This function assumes that the file has been seeked
        # to the correct starting location.
        orig_seek = file.tell()
        struct_bytes = file.read(struct.calcsize(self.StructString))
        file.seek(orig_seek)

        # Load this object with the contents of the data.
        (
            self.StatusBits,
            self.ReservedByte1,
            self.ReservedByte2,
            self.ReservedByte3,
            self.ReservedUint32,
            self.Lba,
            self.Offset,
            self.Length,
            self.RelativeOffset,
        ) = struct.unpack(self.StructString, struct_bytes)

        return self

    def serialize(self) -> bytes:
        r"""Serializes the EFI_FAULT_TOLERANT_WRITE_RECORD.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        return struct.pack(
            self.StructString,
            self.StatusBits,
            self.ReservedByte1,
            self.ReservedByte2,
            self.ReservedByte3,
            self.ReservedUint32,
            self.Lba,
            self.Offset,
            self.Length,
            self.RelativeOffset,
        )



# @file path_utilities.py
# Code to help convert Edk2, absolute, and relative file paths
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
r"""A module for managing Edk2 file paths agnostic to OS path separators ("/" vs "\").

This module converts all windows style paths to Posix file paths internally, but will return
the OS specific path with the exception of of any function that returns an Edk2 style path,
which will always return Posix form.
"""

import concurrent.futures
import errno
import logging
import os
import timeit
from pathlib import Path
from typing import Iterable, Optional


class Edk2Path(object):
    """Represents edk2 file paths.

    Class that helps perform path operations within an EDK workspace.

    Attributes:
        WorkspacePath (str): Absolute path to the workspace root.
        PackagePathList (List[str]): List of absolute paths to a package.

    Attributes are initialized by the constructor and are read-only.

    !!! warning
        Edk2Path performs expensive packages path and package validation when
        instantiated. If using the same Workspace root and packages path, it is
        suggested that only a single Edk2Path instance is instantiated and
        passed to any consumers.
    """

    def __init__(
        self, ws: os.PathLike, package_path_list: Iterable[os.PathLike], error_on_invalid_pp: bool = True
    ) -> "Edk2Path":
        """Constructor.

        Args:
            ws: absolute path or cwd relative path of the workspace.
            package_path_list: list of packages path. Entries can be Absolute path, workspace relative path, or CWD
                relative.
            error_on_invalid_pp: default value is True. If packages path value is invalid raise exception.

        Raises:
            (NotADirectoryError): Invalid workspace or package path directory.
        """
        self.logger = logging.getLogger("Edk2Path")

        # Other code is dependent the following types, so keep it that way:
        #   - self.PackagePathList: List[str]
        #   - self.WorkspacePath: str
        ws = ws.replace("\\", "/")
        workspace_candidate_path = Path(ws)

        if not workspace_candidate_path.is_absolute():
            workspace_candidate_path = Path.cwd() / ws

        if not workspace_candidate_path.is_dir():
            raise NotADirectoryError(errno.ENOENT, os.strerror(errno.ENOENT), workspace_candidate_path.resolve())

        self._workspace_path = workspace_candidate_path

        candidate_package_path_list = []
        start_time = timeit.default_timer()
        for a in [Path(path.replace("\\", "/")) for path in package_path_list]:
            if a.is_absolute():
                candidate_package_path_list.append(a)
            else:
                wsr = self._workspace_path / a
                if wsr.is_dir():
                    candidate_package_path_list.append(wsr)
                else:
                    # assume current working dir relative.  Will catch invalid dir when checking whole list
                    candidate_package_path_list.append(Path.cwd() / a)
        end_time = timeit.default_timer()
        self.logger.log(logging.DEBUG, f"Time to build candidate_package_path_list: {end_time - start_time}")

        start_time = timeit.default_timer()
        invalid_pp = []
        for a in candidate_package_path_list[:]:
            if not a.is_dir():
                self.logger.log(
                    logging.ERROR if error_on_invalid_pp else logging.WARNING,
                    f"Invalid package path entry {a.resolve()}",
                )
                candidate_package_path_list.remove(a)
                invalid_pp.append(str(a.resolve()))
        end_time = timeit.default_timer()
        self.logger.log(logging.DEBUG, f"Time to check package paths: {end_time - start_time}")

        self._package_path_list = candidate_package_path_list

        if invalid_pp and error_on_invalid_pp:
            raise NotADirectoryError(errno.ENOENT, os.strerror(errno.ENOENT), invalid_pp)

        #
        # Nested package check - ensure packages do not exist in a linear
        # path hierarchy.
        #
        # 1. Build a dictionary for each package path.
        #      - Key = Package path
        #      - Value = List of packages discovered in package path
        # 2. Enumerate all keys in dictionary checking if any package
        #    is relative (nested) to each other.
        # 3. Raise an Exception if two packages are found to be nested.
        #
        if "PYTOOL_TEMPORARILY_IGNORE_NESTED_EDK_PACKAGES" in os.environ:
            warning = (
                "PYTOOL_TEMPORARILY_IGNORE_NESTED_EDK_PACKAGES is no longer used by edk2-pytool-library, but is "
                "detected in your environment. Please remove this environment variable."
            )
            self.logger.log(logging.WARNING, warning)
        if "PYTOOL_IGNORE_KNOWN_BAD_NESTED_PACKAGES" == os.environ:
            warning = (
                "PYTOOL_IGNORE_KNOWN_BAD_NESTED_PACKAGES is no longer used by edk2-pytool-library, but is "
                "detected in your environment. Please remove this environment variable."
            )
            self.logger.log(logging.WARNING, warning)

        package_path_packages = {}

        start_time = timeit.default_timer()

        def get_paths(package_path):
            paths = []
            for root, _, files in os.walk(package_path):
                for file in files:
                    if file.endswith(".dec"):
                        paths.append(Path(root))
            return package_path, paths

        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = executor.map(get_paths, self._package_path_list)

        for package_path, paths in results:
            package_path_packages[package_path] = paths

        end_time = timeit.default_timer()
        self.logger.log(logging.DEBUG, f"Time to build package_path_packages: {end_time - start_time}")

        start_time = timeit.default_timer()
        for package_path, packages in package_path_packages.items():
            for i, package in enumerate(packages):
                for j in range(i + 1, len(packages)):
                    comp_package = packages[j]
                    if package.is_relative_to(comp_package) or comp_package.is_relative_to(package):
                        self.logger.log(
                            logging.DEBUG,
                            f"[{str(package)}] and [{str(comp_package)}] are nested. Nested packages are not allowed "
                            "and may result in incorrect conversions from absolute path to edk2 package path relative "
                            "paths.",
                        )
        end_time = timeit.default_timer()
        self.logger.log(logging.DEBUG, f"Time to check nested packages: {end_time - start_time}")

    @property
    def WorkspacePath(self: "Edk2Path") -> str:
        """Workspace Path as a string."""
        return str(self._workspace_path)

    @property
    def PackagePathList(self: "Edk2Path") -> list[str]:
        """List of package paths as strings."""
        return [str(p) for p in self._package_path_list]

    def GetEdk2RelativePathFromAbsolutePath(self, *abspath: str) -> str:
        """Transforms an absolute path to an edk2 path relative to the workspace or a packages path.

        Args:
            *abspath: absolute path to a file or directory. Can be the entire path or parts of the path provided
                separately. Supports both Windows and Posix style paths

        Returns:
            (str): POSIX-like path relative to the workspace or packages path
            (None): abspath is None
            (None): path is not valid

        Example:
            ```python
                rel_path = edk2path.GetEdk2RelativePathFromAbsolutePath("C:/Workspace/edk2/MdePkg/Include")
                rel_path = edk2path.GetEdk2RelativePathFromAbsolutePath("C:/Workspace", "edk2", "MdePkg", "Include")
            ```
        """
        if abspath == (None,):
            return None

        abspath = Path(*[part.replace("\\", "/") for part in abspath])

        relpath = None
        found = False

        # Check if the Absolute path starts with any of the package paths. If a match is found, build the relative
        # path based off that package.
        #
        # Sort the package paths from from longest to shortest. This handles the case where a package and a package
        # path are in the same directory. See the following path_utilities_test for a detailed explanation of the
        # scenario: test_get_relative_path_when_folder_is_next_to_package
        for packagepath in sorted(self._package_path_list, reverse=True):
            # If a match is found, use the original string to avoid change in case
            if abspath.is_relative_to(packagepath):
                self.logger.debug("Successfully converted AbsPath to Edk2Relative Path using PackagePath")
                relpath = abspath.relative_to(packagepath)
                found = True
                break

        # If a match was not found, check if absolute path is based on the workspace root.
        if not found and abspath.is_relative_to(self._workspace_path):
            self.logger.debug("Successfully converted AbsPath to Edk2Relative Path using WorkspacePath")
            relpath = abspath.relative_to(self._workspace_path)
            found = True

        if found:
            relpath = relpath.as_posix()
            self.logger.debug(f"[{abspath}] -> [{relpath}]")
            return relpath

        # Absolute path was not in reference to a package path or the workspace root.
        self.logger.error("Failed to convert AbsPath to Edk2Relative Path")
        self.logger.error(f"AbsolutePath: {abspath}")
        return None

    def GetAbsolutePathOnThisSystemFromEdk2RelativePath(self, *relpath: str, log_errors: Optional[bool] = True) -> str:
        """Given a relative path return an absolute path to the file in this workspace.

        Args:
            *relpath: Relative path to convert. Can be the entire path or parts of the path provided separately
            log_errors: whether to log errors

        Returns:
            (str): absolute path in the OS specific form
            (None): invalid relpath
            (None): Unable to get the absolute path

        Example:
            ```python
                abs_path = edk2path.GetAbsolutePathOnThisSystemFromEdk2RelativePath("MdePkg/Include")
                abs_path = edk2path.GetAbsolutePathOnThisSystemFromEdk2RelativePath("MdePkg", "Include")
            ```
        """
        if relpath == (None,):
            return None

        relpath = Path(*[part.replace("\\", "/") for part in relpath])
        abspath = self._workspace_path / relpath
        if abspath.exists():
            return str(abspath)

        for a in self._package_path_list:
            abspath = a / relpath
            if abspath.exists():
                return str(abspath)
        if log_errors:
            self.logger.error("Failed to convert Edk2Relative Path to an Absolute Path on this system.")
            self.logger.error("Relative Path: %s" % relpath)

        return None

    def GetContainingPackage(self, InputPath: str) -> str:
        """Finds the package that contains the given path.

        This isn't perfect, but at least identifies the direcotry consistently.


        Args:
            InputPath: absolute path to a file, directory, or module. Supports both windows and posix paths.

        Returns:
            (str): name of the package that the path is in.
        """
        self.logger.debug("GetContainingPackage: %s" % InputPath)
        InputPath = Path(InputPath.replace("\\", "/"))
        # Make a list that has the path case normalized for comparison.
        # Note: This only does anything on Windows

        # 1. Handle the case that InputPath is not in the workspace tree
        path_root = None
        if not InputPath.is_relative_to(self._workspace_path):
            for p in self._package_path_list:
                if InputPath.is_relative_to(p):
                    path_root = p
                    break
            if not path_root:
                return None
        else:
            path_root = self._workspace_path

        # 2. Determine if the path is under a package in the workspace

        # Start the search within the first available directory. If provided InputPath is a directory, start there,
        # else (if InputPath is a file) move to it's parent directory and start there.
        if InputPath.is_dir():
            dirpath = InputPath
        else:
            dirpath = InputPath.parent

        while path_root != dirpath:
            if dirpath.exists():
                for f in dirpath.iterdir():
                    if f.suffix.lower() == ".dec":
                        return dirpath.name

            dirpath = dirpath.parent

        return None

    def GetContainingModules(self, input_path: str) -> list[str]:
        """Find the list of modules (inf files) for a file path.

        Note: This function only accepts absolute paths. An exception will
              be raised if a non-absolute path is given.

        Note: If input_path does not exist in the filesystem, this function
              will try to return the likely containing module(s) but if the
              entire module has been deleted, this isn't possible.

        - If a .inf file is given, that file is returned.
        - Otherwise, the nearest set of .inf files (in the closest parent)
          will be returned in a list of file path strings.

        Args:
            input_path: Absolute path to a file, directory, or module.
                Supports both Windows and Posix like paths.

        Returns:
            (list[str]): Absolute paths of .inf files that could be the
                         containing module.
        """
        input_path = Path(input_path.replace("\\", "/"))
        if not input_path.is_absolute():
            # Todo: Return a more specific exception type when
            # https://github.com/tianocore/edk2-pytool-library/issues/184 is
            # implemented.
            raise Exception("Module path must be absolute.")

        all_root_paths = self._package_path_list + [self._workspace_path]

        # For each root path, find the maximum allowed root in its hierarchy.
        maximum_root_paths = all_root_paths
        for root_path in maximum_root_paths:
            for other_root_path in maximum_root_paths[:]:
                if root_path == other_root_path:
                    continue
                if root_path.is_relative_to(other_root_path):
                    if len(root_path.parts) > len(other_root_path.parts):
                        maximum_root_paths.remove(root_path)
                    else:
                        maximum_root_paths.remove(other_root_path)

        # Verify the file path is within a valid workspace or package path
        # directory.
        for path in maximum_root_paths:
            if input_path.is_relative_to(path):
                break
        else:
            return []

        modules = []
        if input_path.suffix.lower() == ".inf":
            # Return the file path given since it is a module .inf file
            modules = [str(input_path)]

        if not modules:
            # Continue to ascend directories up to a maximum root path.
            #
            # This handles cases like:
            #   ModuleDir/      |   ModuleDir/      | ...similarly nested files
            #     Module.inf    |     Module.inf    |
            #     x64/          |     Common/       |
            #       file.c      |       X64/        |
            #                   |         file.c    |
            #
            # The terminating condition of the loop is when a maximum root
            # path has been reached.
            #
            # A maximum root path represents the maximum allowed ascension
            # point in the input_path directory hierarchy as sub-roots like
            # a package path pointing under a workspace path are already
            # accounted for during maximum root path filtering.
            #
            # Given a root path is either the workspace or a package path,
            # neither of which are a module directory, once that point is
            # reached, all possible module candidates are exhausted.
            current_dir = input_path.parent
            while current_dir not in maximum_root_paths:
                if current_dir.is_dir():
                    current_dir_inf_files = [
                        str(f) for f in current_dir.iterdir() if f.is_file() and f.suffix.lower() == ".inf"
                    ]
                    if current_dir_inf_files:
                        # A .inf file(s) exist in this directory.
                        #
                        # Since this is the closest parent that can be considered
                        # a module, return the .inf files as module candidates.
                        modules.extend(current_dir_inf_files)
                        break

                current_dir = current_dir.parent

        return modules



## @file
# Module that encodes and decodes a FMP_PAYLOAD_HEADER with a payload.
# The FMP_PAYLOAD_HEADER is processed by the FmpPayloadHeaderLib in the
# FmpDevicePkg.
#
# Copyright (c) 2018, Intel Corporation. All rights reserved.<BR>
# SPDX-License-Identifier: BSD-2-Clause-Patent
#

"""Module that encodes and decodes a FMP_PAYLOAD_HEADER with a payload.

The FMP_PAYLOAD_HEADER is processed by the FmpPayloadHeaderLib in the
FmpDevicePkg.
"""

import struct


def _SIGNATURE_32(A: str, B: str, C: str, D: str) -> bytes:
    return struct.unpack("=I", bytearray(A + B + C + D, "ascii"))[0]


def _SIGNATURE_32_TO_STRING(Signature: int) -> bytes:
    return struct.pack("<I", Signature).decode()


class FmpPayloadHeaderClass(object):
    """An object representing a FMP_PAYLOAD_HEADER.

    typedef struct {
        UINT32  Signature;
        UINT32  HeaderSize;
        UINT32  FwVersion;
        UINT32  LowestSupportedVersion;
    } FMP_PAYLOAD_HEADER;

    #define FMP_PAYLOAD_HEADER_SIGNATURE SIGNATURE_32 ('M', 'S', 'S', '1')
    """

    _StructFormat = "<IIII"
    _StructSize = struct.calcsize(_StructFormat)

    _FMP_PAYLOAD_HEADER_SIGNATURE = _SIGNATURE_32("M", "S", "S", "1")

    def __init__(self) -> "FmpPayloadHeaderClass":
        """Inits an empty object."""
        self.Signature = self._FMP_PAYLOAD_HEADER_SIGNATURE
        self.HeaderSize = self._StructSize
        self.FwVersion = 0x00000000
        self.LowestSupportedVersion = 0x00000000
        self.Payload = b""

    def Encode(self) -> bytes:
        r"""Serializes the Header + payload.

        Returns:
            (bytes): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        FmpPayloadHeader = struct.pack(
            self._StructFormat, self.Signature, self.HeaderSize, self.FwVersion, self.LowestSupportedVersion
        )
        return FmpPayloadHeader + self.Payload

    def Decode(self, Buffer: bytes) -> bytes:
        """Loads data into the Object by parsing a buffer.

        Args:
            Buffer (obj): Buffer containing the data

        Returns:
            (str): string of binary representing the payload

        Raises:
            (ValueError): Invalid Buffer
            (ValueError): Invalid Signature
            (ValueError): Invalid Header size
        """
        if len(Buffer) < self._StructSize:
            raise ValueError
        (Signature, HeaderSize, FwVersion, LowestSupportedVersion) = struct.unpack(
            self._StructFormat, Buffer[0 : self._StructSize]
        )
        if Signature != self._FMP_PAYLOAD_HEADER_SIGNATURE:
            raise ValueError
        if HeaderSize < self._StructSize:
            raise ValueError
        self.Signature = Signature
        self.HeaderSize = HeaderSize
        self.FwVersion = FwVersion
        self.LowestSupportedVersion = LowestSupportedVersion
        self.Payload = Buffer[self.HeaderSize :]

        return self.Payload

    def DumpInfo(self) -> None:
        """Prints payload header information."""
        print(
            "FMP_PAYLOAD_HEADER.Signature              = {Signature:08X} ({SignatureString})".format(
                Signature=self.Signature, SignatureString=_SIGNATURE_32_TO_STRING(self.Signature)
            )
        )
        print("FMP_PAYLOAD_HEADER.HeaderSize             = {HeaderSize:08X}".format(HeaderSize=self.HeaderSize))
        print("FMP_PAYLOAD_HEADER.FwVersion              = {FwVersion:08X}".format(FwVersion=self.FwVersion))
        print(
            "FMP_PAYLOAD_HEADER.LowestSupportedVersion = {LowestSupportedVersion:08X}".format(
                LowestSupportedVersion=self.LowestSupportedVersion
            )
        )
        print("sizeof (Payload)                          = {Size:08X}".format(Size=len(self.Payload)))



# @file instaced_fv.py
# A module to run a table generator that uses a fdf and environment information to generate a table of information
# about instanced fvs where each row is a unique fv.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A module to generate a table containing fv information."""

import logging
import re
from pathlib import Path
from typing import Any

from edk2toollib.database import Fv, InstancedInf, Session
from edk2toollib.database.tables import TableGenerator
from edk2toollib.uefi.edk2.parsers.fdf_parser import FdfParser as FdfP
from edk2toollib.uefi.edk2.path_utilities import Edk2Path


class InstancedFvTable(TableGenerator):
    """A Table Generator that parses a single FDF file and generates a table containing FV information.

    !!! warning
        This table generator relies on the instanced_inf_table generator to be run first.
    """  # noqa: E501

    INFOPTS = re.compile(r"(RuleOverride|file_guid|version|ui|use)\s*=.+\s+(.+\.inf)", re.IGNORECASE)

    def __init__(self, *args: Any, **kwargs: Any) -> "InstancedFvTable":
        """Initialize the query with the specific settings."""

    def parse(self, session: Session, pathobj: Edk2Path, env_id: str, env: dict) -> None:
        """Parse the workspace and update the database."""
        self.pathobj = pathobj
        self.ws = Path(self.pathobj.WorkspacePath)
        self.env = env
        self.env_id = env_id
        self.dsc = self.env.get("ACTIVE_PLATFORM", None)
        self.fdf = self.env.get("FLASH_DEFINITION", None)
        self.arch = self.env["TARGET_ARCH"].split(" ")
        self.target = self.env["TARGET"]

        if self.dsc is None or self.fdf is None:
            logging.debug("DSC or FDF not found in environment. Skipping InstancedFvTable")
            return

        # Our DscParser subclass can now parse components, their scope, and their overrides
        fdfp = FdfP().SetEdk2Path(self.pathobj)
        fdfp.SetInputVars(self.env)
        fdfp.ParseFile(self.fdf)

        all_components = {inf.path: inf for inf in session.query(InstancedInf).filter_by(env=env_id, cls=None).all()}
        for fv in fdfp.FVs:
            inf_list = []  # Some INF's have extra options. We only need the INF
            for inf in fdfp.FVs[fv]["Infs"]:
                options = InstancedFvTable.INFOPTS.findall(inf)
                if len(options) > 0:
                    inf = options[0][1]

                # Convert to absolute, and back to relative to ensure we get the closest pp relative path
                # i.e. if we have two package paths: ("MyPP", and "MyPP/Subfolder"), in the FDF, devs
                # can specify INFs are either ("Subfolder/MyPkg/../MyPkg.inf" or "MyPkg/../MyPkg.inf")
                # However in the database, we want the closest match, i.e. "MyPkg/../MyPkg.inf", even if
                # they are providing ("Subfolder/MyPkg/../MyPkg.inf"). "GetEdk2RelativePathFromAbsolutePath"
                # always returns the relative path from the closest package path.
                inf = self.pathobj.GetAbsolutePathOnThisSystemFromEdk2RelativePath(inf)
                inf = self.pathobj.GetEdk2RelativePathFromAbsolutePath(inf)
                inf_list.append(Path(inf).as_posix())

            filtered = []
            for inf in inf_list:
                if inf not in all_components:
                    logging.warning(f"INF [{inf}] not found in database.")
                else:
                    filtered.append(inf)

            fv = Fv(env=env_id, name=fv, fdf=self.fdf, infs=[all_components.get(inf) for inf in filtered])
            session.add(fv)
            session.commit()



## @file
# Script to generate Cat files for capsule update based on supplied inf file.
# This uses the winsdk and the command line tool Inf2Cat.exe
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Script to generate Cat files for capsule update.

Based on a supplied inf file and uses the winsdk and command line tool Inf2Cat.exe
"""

import logging
import os
from typing import Optional

from edk2toollib.utility_functions import RunCmd
from edk2toollib.windows.locate_tools import FindToolInWinSdk


class CatGenerator(object):
    """A cat file generator.

    Attributes:
        arch (str): a supported architecture
        os (str): a supported os
    """

    SUPPORTED_OS = {
        "win10": "10",
        "10": "10",
        "10_au": "10_AU",
        "10_rs2": "10_RS2",
        "10_rs3": "10_RS3",
        "10_rs4": "10_RS4",
        "server10": "Server10",
        "server2016": "Server2016",
        "serverrs2": "ServerRS2",
        "serverrs3": "ServerRS3",
        "serverrs4": "ServerRS4",
    }

    def __init__(self, arch: str, os: str) -> "CatGenerator":
        """Inits a Cat Generator.

        Args:
            arch (str): a supported arch
            os (str): a supported os
        """
        self.Arch = arch
        self.OperatingSystem = os

    @property
    def Arch(self) -> str:
        """Returns the attribute arch."""
        return self._arch

    @Arch.setter
    def Arch(self, value: str) -> None:
        """Validates the arch before setting it.

        Raises:
            (ValueError): Invalid Architecture
        """
        value = value.lower()
        if (value == "x64") or (value == "amd64"):  # support amd64 value so INF and CAT tools can use same arch value
            self._arch = "X64"
        elif value == "arm":
            self._arch = "ARM"
        elif (value == "arm64") or (value == "aarch64"):  # support UEFI defined aarch64 value as well
            self._arch = "ARM64"
        else:
            logging.critical("Unsupported Architecture: %s", value)
            raise ValueError("Unsupported Architecture")

    @property
    def OperatingSystem(self) -> str:
        """Returns the Operating system attribute."""
        return self._operatingsystem

    @OperatingSystem.setter
    def OperatingSystem(self, value: str) -> None:
        """Validates the OS is supported before setting the attribute.

        Raises:
            (ValueError): Operating system is unsupported
        """
        key = value.lower()
        if key not in CatGenerator.SUPPORTED_OS.keys():
            logging.critical("Unsupported Operating System: %s", key)
            raise ValueError("Unsupported Operating System")
        self._operatingsystem = CatGenerator.SUPPORTED_OS[key]

    def MakeCat(self, OutputCatFile: str, PathToInf2CatTool: Optional[str] = None) -> int:
        """Generates a cat file to the outputcatfile directory.

        Args:
            OutputCatFile (str): Where to place the output cat file.
            PathToInf2CatTool (:obj:`str`, optional): path to Inf2CatTool if known.

        Raises:
            (Exception): Invalid Inf2CatTool path or unable to find it.
            (Exception): Inf2CatTool failed
            (Exception): Cat file not found, but tool executed successfully
        """
        # Find Inf2Cat tool
        if PathToInf2CatTool is None:
            PathToInf2CatTool = FindToolInWinSdk("Inf2Cat.exe")
        # check if exists
        if PathToInf2CatTool is None or not os.path.exists(PathToInf2CatTool):
            raise Exception(
                "Can't find Inf2Cat on this machine.  Please install the Windows 10 WDK - "
                "https://developer.microsoft.com/en-us/windows/hardware/windows-driver-kit"
            )

        OutputFolder = os.path.dirname(OutputCatFile)
        # Make Cat file
        cmd = "/driver:. /os:" + self.OperatingSystem + "_" + self.Arch + " /verbose /uselocaltime"
        ret = RunCmd(PathToInf2CatTool, cmd, workingdir=OutputFolder)
        if ret != 0:
            raise Exception("Creating Cat file Failed with errorcode %d" % ret)
        if not os.path.isfile(OutputCatFile):
            raise Exception("CAT file (%s) not created" % OutputCatFile)

        return 0



# @file dsc.py
# Data model for the EDK II DSC
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Data model for the EDK II DSC."""

# There will be some overlap between the objects for DSC files and FDF files
import collections
from typing import Any, Optional

DEFAULT_SECTION_TYPE = "COMMON"


class dsc_set(set):
    """a DSC set object."""

    def __init__(self, allowed_classes: list = None) -> "dsc_set":
        """Initializes an empty set."""
        if allowed_classes is None:
            allowed_classes = []
        self._allowed_classes = set(allowed_classes)

    def add(self, item: Any) -> None:  # noqa: ANN401
        """Adds the item to the set.

        Raises:
            (ValueError): if the type is not allowed to be added.
        """
        if len(self._allowed_classes) > 0 and type(item) not in self._allowed_classes:
            raise ValueError(f"Cannot add {type(item)} to restricted set: {self._allowed_classes}")
        if item in self:
            super().discard(item)
            # NEXTVER: get add the old_item to item
        super().add(item)


class dsc_dict(collections.OrderedDict):
    """A dictionary that allows specific classes as headers and sections."""

    def __init__(
        self,
        allowed_key_classes: Optional[str] = None,
        allowed_value_classes: Optional[Any] = None,  # noqa: ANN401
    ) -> "dsc_dict":
        """Initializes a dsc_dict."""
        if allowed_key_classes is None:
            allowed_key_classes = []
        if allowed_value_classes is None:
            allowed_value_classes = []
        self._allowed_key_classes = set(allowed_key_classes)
        self._allowed_value_classes = set(allowed_value_classes)

    def __missing__(self, key: int) -> str:
        """Describes how to handle a missing key.

        Makes a new dsc_set if allowed.

        Raises:
            (KeyError): If not allowed
        """
        if len(self._allowed_value_classes) > 0:  # if we have specified allowed value classes, make a new dsc_set
            self[key] = dsc_set(self._allowed_value_classes)
            return self[key]
        raise KeyError(key)

    def __setitem__(self, key: int, val: str) -> None:
        """Setting a key value pair in the Ordered Dictionary.

        Raises:
            (ValueError): Adding a invalid key to restricted set
            (ValueError): Adding a invalid set to a restricted dict
            (ValueError): Adding a section that already exists
        """
        if len(self._allowed_key_classes) > 0 and type(key) not in self._allowed_key_classes:
            raise ValueError(f"Cannot add {type(key)} to restricted set: {self._allowed_key_classes}")

        if len(self._allowed_value_classes) > 0:
            if isinstance(val, set) and len(val) == 0:  # if it's an empty set, convert it to a dsc_set
                val = dsc_set(allowed_classes=self._allowed_value_classes)
            if isinstance(val, dsc_set):
                if val._allowed_classes != self._allowed_value_classes:
                    raise ValueError(
                        f"Cannot add set:{val._allowed_classes} to restricted dict: {self._allowed_value_classes}"
                    )
            elif type(val) not in self._allowed_value_classes:
                raise ValueError(f"Cannot add {type(val)} to restricted dict: {self._allowed_value_classes}")
        if key in self:
            # NEXTVER merge these together?
            raise ValueError(f"Cannot add section {key} since it already exists")
        dict.__setitem__(self, key, val)


class dsc:
    """Class representing a DSC."""

    def __init__(self, file_path: Optional[str] = None) -> "dsc":
        """Inits dsc type."""
        # The EDK2 path to this particular DSC, if is is None it means it was created from a stream and has no file
        self.file_path = file_path
        # parameters added for clarity
        self.skus = dsc_set(allowed_classes=[definition, sku_id])  # this is a set of SKUs
        self.components = dsc_dict(
            allowed_key_classes=[
                dsc_section_type,
            ],
            allowed_value_classes=[component, definition],
        )
        self.libraries = dsc_dict(
            allowed_key_classes=[
                dsc_section_type,
            ],
            allowed_value_classes=[library, definition],
        )
        self.library_classes = dsc_dict(
            allowed_key_classes=[
                dsc_section_type,
            ],
            allowed_value_classes=[library_class, definition],
        )
        self.build_options = dsc_dict(
            allowed_key_classes=[
                dsc_buildoption_section_type,
            ],
            allowed_value_classes=[build_option, definition],
        )
        self.pcds = dsc_dict(
            allowed_key_classes=[
                dsc_pcd_section_type,
            ],
            allowed_value_classes=[pcd, pcd_typed, pcd_variable],
        )
        self.defines = dsc_set(
            allowed_classes=[
                definition,
            ]
        )
        self.default_stores = dsc_set(allowed_classes=[definition, default_store])

        # NEXTVER: should we populate the default information into the DSC object?
        # FOR EXAMPLE: default_stores and skus

    def __eq__(self, other: "dsc") -> bool:
        """Enables equality comparisons (a == b).

        Warning:
            This doesn't check for a perfect copy of everything
            this is mainly focused on does it define the same things
        """
        if type(other) is not dsc:
            return False
        if other.skus != self.skus:
            return False
        if other.components != self.components:
            return False
        if other.library_classes != self.library_classes:
            return False
        if other.build_options != self.build_options:
            return False
        if other.pcds != self.pcds:
            return False
        if other.defines != self.defines:
            return False
        return True


class dsc_section_type:
    """dsc section type.

    Attributes:
        arch (str): architecture
        module_type (str): module type
    """

    dsc_module_types = [
        "COMMON",
        "BASE",
        "SEC",
        "PEI_CORE",
        "PEIM",
        "DXE_CORE",
        "DXE_DRIVER",
        "DXE_RUNTIME_DRIVER",
        "DXE_SAL_DRIVER",
        "DXE_SMM_DRIVER",
        "SMM_CORE",
        "UEFI_DRIVER",
        "UEFI_APPLICATION",
        "USER_DEFINED",
    ]

    def __init__(self, arch: str = "common", module_type: str = "common") -> "dsc_section_type":
        """Inits dsc section type."""
        self.arch = arch.upper().strip()
        self.module_type = module_type.upper().strip()
        if not dsc_section_type.IsValidModuleType(self.module_type):
            raise ValueError(f"{module_type} is not a proper module type for dsc section")

    def __hash__(self) -> int:
        """Returns the hash of an object for hashtables."""
        arch = "*" if (self.arch == "COMMON" or self.arch == "DEFAULT") else self.arch
        return hash((arch, self.module_type))

    def __eq__(self, other: "dsc_section_type") -> bool:
        """Enables equality comparisons (a == b)."""
        if type(other) is not dsc_section_type:
            return False
        arch = "*" if (self.arch == "COMMON" or self.arch == "DEFAULT") else self.arch
        arch2 = "*" if (other.arch == "COMMON" or other.arch == "DEFAULT") else other.arch
        return self.module_type == other.module_type or arch == arch2

    def __repr__(self) -> str:
        """A string representation of the object."""
        attributes = f".{self.arch}.{self.module_type}"
        return attributes

    @classmethod
    def IsValidModuleType(cls: "dsc_section_type", name: str) -> bool:
        """If the module type is valid or not."""
        return name in cls.dsc_module_types


class dsc_buildoption_section_type(dsc_section_type):
    """Build option section type.

    Attributes:
        codebase (:obj:`str`, optional): the codebase type

    Examples:
        ```
        BuildOptions.$(arch).CodeBase.Edk2ModuleType
        BuildOptions.$(arch).CodeBase
        BuildOptions.common.CodeBase
        BuildOptions.$(arch)
        BuildOptions.common
        BuildOptions
        ```
    """

    def __init__(
        self, arch: str = "common", codebase: str = "common", module_type: str = "common"
    ) -> "dsc_buildoption_section_type":
        """Inits dsc build option section type."""
        super().__init__(arch, module_type)
        self.codebase = codebase.upper().strip()
        if not self.IsValidCodeBase(self.codebase):
            raise ValueError(f"{codebase} is not a valid codebase type")

    @classmethod
    def IsValidCodeBase(cls: "dsc_buildoption_section_type", codebase: str) -> bool:
        """Determines if codebase is valid.

        Args:
            codebase: codebase

        Returns:
            (bool): True if COMMON, EDK, EDKII else false
        """
        return codebase in ["COMMON", "EDK", "EDKII"]

    def __hash__(self) -> int:
        """Returns the hash of an object for hashtables."""
        return hash((super().__hash__(), self.codebase))

    def __eq__(self, other: "dsc_buildoption_section_type") -> bool:
        """Enables equality comparisons (a == b)."""
        if type(other) is not dsc_buildoption_section_type:
            return False
        if not (other.module_type == self.module_type and other.codebase == self.codebase):
            return False
        return other.module_type == self.module_type

    def __repr__(self) -> str:
        """A string representation of the object."""
        return f"{self.arch}.{self.codebase}.{self.module_type}"


dsc_pcd_types = [
    "FEATUREFLAG",
    "PATCHABLEINMODULE",
    "FIXEDATBUILD",
    "DYNAMIC",
    "DYNAMICEX",
    "DYNAMICDEFAULT",
    "DYNAMICHII",
    "DYNAMICVPD",
    "DYNAMICEXHII",
    "DYNAMICEXVPD",
]


class dsc_pcd_section_type:
    """This class is uses to define the PCD section type inside a component.

    Attributes:
        pcdtype: pcd type
        arch: defaults to common
        sku: defaults to DEFAULT
        store: if none, we don't have anything done
    """

    def __init__(
        self, pcdtype: str, arch: str = "common", sku: str = "DEFAULT", store: Optional[str] = None
    ) -> "dsc_pcd_section_type":
        """Inits the dsc_pcd_section object."""
        # if store is none, then we don't have anything done
        self.arch = arch.upper().strip()
        self.pcd_type = pcdtype.upper().strip()
        self.default_store = None if store is None else store.strip()
        if self.pcd_type not in dsc_pcd_types:
            raise ValueError(f"{pcdtype} is not a proper PCD type")
        if not self.pcd_type.endswith("HII") and self.default_store is not None:
            raise ValueError(f"{pcdtype} does not allow for a store to be specified")
        self.sku = sku.upper()

    def __hash__(self) -> int:
        """Returns the hash of an object for hashtables."""
        return hash((self.arch, self.pcd_type))

    def __repr__(self) -> str:
        """A string representation of the object."""
        store = "" if self.default_store is None else f".{self.default_store}"
        return f"Pcds{self.pcd_type}.{self.arch}.{self.sku}{store}"

    def __eq__(self, other: "dsc_pcd_section_type") -> bool:
        """Enables equality comparisons (a == b)."""
        if type(other) is not dsc_pcd_section_type:
            return False
        return self.pcd_type == other.pcd_type and self.arch == other.arch


class dsc_pcd_component_type(dsc_pcd_section_type):
    """This class is uses to define the PCD type inside a component."""

    def __init__(self, pcdtype: str) -> "dsc_pcd_component_type":
        """Inits the dsc_pcd_component object."""
        super().__init__(pcdtype)

    def __repr__(self) -> str:
        """A string representation of the object."""
        return f"Pcds{self.pcd_type}"

    def __hash__(self) -> int:
        """Returns the hash of an object for hashtables."""
        return hash(self.pcd_type)

    def __eq__(self, other: "dsc_pcd_component_type") -> bool:
        """Enables equality comparisons (a == b)."""
        if type(other) is not dsc_pcd_component_type:
            return False
        return self.pcd_type == other.pcd_type


class sku_id:
    """Contains the data for a sku.

    Args:
        id: sku id number
        name: name of the sku
        parent: parent
        source_info: source info
    """

    def __init__(
        self, id: int = 0, name: str = "DEFAULT", parent: str = "DEFAULT", source_info: Optional["source_info"] = None
    ) -> "sku_id":
        """Inits the sku_id object."""
        self.id = id
        self.name = name
        self.parent = parent  # the default parent is default
        self.source_info = source_info

    def __eq__(self, other: "sku_id") -> bool:
        """Enables equality comparisons (a == b)."""
        if type(other) is not sku_id:
            return False
        return self.id == other.id or self.name == other.name

    def __hash__(self) -> int:
        """Returns the hash of an object for hashtables."""
        # we return zero because we want all the skus to hash to the same bucket
        # this won't be performant for large numbers of skus, which hopefully won't happen
        # we instead rely on __eq__ since we want to collide on two different attributes
        # since we want to make sure names and id's are unique
        return 0

    def __repr__(self) -> str:
        """A string representation of the object."""
        return f"{self.id}|{self.name}|{self.parent}"


class component:
    """Contains the data for a component for the EDK build system to build."""

    def __init__(self, inf: str, source_info: Optional["source_info"] = None) -> "component":
        """Inits the component object."""
        self.library_classes = set()  # a list of libraries that this component uses
        self.pcds = {}  # a dictionary of PCD's that are keyed by dsc_pcd_component_type, they are sets
        self.defines = set()  # a set of defines
        self.build_options = set()  # a set of build options for this component
        self.inf = inf  # the EDK2 relative path to the source INF
        self.source_info = source_info

    def __eq__(self, other: "component") -> bool:
        """Enables equality comparisons (a == b)."""
        if type(other) is not component:
            return False
        return self.inf == other.inf  # NEXTVER: should this be case insensitive?

    def __hash__(self) -> int:
        """Returns the hash of an object for hashtables."""
        return hash(self.inf)

    def __repr__(self) -> str:
        """A string representation of the object."""
        source = str(self.source_info) if source_info is not None else ""
        return f"{self.inf} @ {source}"


class definition:
    """Contains the information on a definition.

    Args:
        name: definition name
        value: definition value
        local: if the value is local or not
        source_info: source info
    """

    def __init__(
        self, name: str, value: str, local: bool = False, source_info: Optional["source_info"] = None
    ) -> "definition":
        """Inits the definition object.

        NOTE: Local means DEFINE is in front and is localized to that particular section
        """
        self.name = name
        self.value = value
        self.local = local
        self.source_info = source_info

    def __repr__(self) -> str:
        """A string representation of the object."""
        string = ""
        if self.local:
            string = "DEFINE "
        string += f"{self.name} = {self.value} @ {self.source_info}"
        return string

    def __hash__(self) -> int:
        """Returns the hash of an object for hashtables."""
        return hash(self.name)

    def __eq__(self, other: "definition") -> bool:
        """Enables equality comparisons (a == b)."""
        if type(other) is not definition:
            return False
        return other.name == self.name


class library:
    """Contains the data for a specific EDK library.

    Attributes:
        inf (str): inf file
        source_info (:obj:`obj`, optional): source info
    """

    def __init__(self, inf: str, source_info: Optional["source_info"] = None) -> "library":
        """Inits the Library object."""
        self.inf = inf
        self.source_info = source_info

    def __eq__(self, other: "library") -> bool:
        """Enables equality comparisons (a == b)."""
        if type(other) is not library:
            return False
        return self.inf == other.inf

    def __hash__(self) -> int:
        """Returns the hash of an object for hashtables."""
        return hash(self.inf)  # NEXTVER how to figure out if they hash to the same spot?

    def __repr__(self) -> str:
        """A string representation of the object."""
        return f"{self.inf} @ {self.source_info}"


class library_class:
    """Contains the data for a specific EDK2 library class.

    Attributes:
        libraryclass: the EDK2 library class
        inf): the inf
        source_info: source info
    """

    def __init__(self, libraryclass: str, inf: str, source_info: Optional["source_info"] = None) -> "library_class":
        """Inits the Library class object.

        Args:
            libraryclass: the EDK2 library class
            inf: the inf
            source_info: source info
        """
        self.libraryclass = libraryclass
        self.inf = inf
        self.source_info = source_info

    def __eq__(self, other: "library_class") -> bool:
        """Enables equality comparisons (a == b)."""
        if type(other) is not library_class:
            return False
        # if they're both null
        if self.libraryclass.lower() == "null" and other.libraryclass.lower() == "null":
            return self.inf == other.inf
        return self.libraryclass.lower() == other.libraryclass.lower()

    def __hash__(self) -> int:
        """Returns the hash of an object for hashtables."""
        # if we're a null lib, we want the hash to be based on the inf path
        if self.libraryclass.lower() == "null":
            return hash(self.inf)
        else:
            return hash(self.libraryclass)

    def __repr__(self) -> str:
        """A string representation of the object."""
        return f"{self.libraryclass}|{self.inf} @ {self.source_info}"


class pcd:
    """A PCD object.

    Attributes:
        namespace: namespace
        name: pcd name
        value: pcd value
        source_info: source info

    EXAMPLE: PcdTokenSpaceGuidCName.PcdCName|Value
    """

    def __init__(self, namespace: str, name: str, value: str, source_info: Optional["source_info"] = None) -> "pcd":
        """Inits a PCD object.

        Args:
            namespace: namespace
            name: pcd name
            value: pcd value
            source_info: source info
        """
        self.namespace = namespace
        self.name = name
        self.value = value
        self.source_info = source_info

    def __eq__(self, other: "pcd") -> bool:
        """Enables equality comparisons (a == b)."""
        if not issubclass(other.__class__, pcd):
            return False
        return self.namespace == other.namespace and self.name == other.name

    def __hash__(self) -> int:
        """Returns the hash of an object for hashtables."""
        return hash(f"{self.namespace}.{self.name}")

    def __repr__(self) -> str:
        """A string representation of the object."""
        return f"{self.namespace}.{self.name} = {self.value} @ {self.source_info}"


class pcd_typed(pcd):
    """A Typed PCD object.

    Attributes:
        datum_type: Data Type
        max_size: max size of the data type

    EXAMPLE: PcdTokenSpaceGuidCName.PcdCName|Value[|DatumType[|MaximumDatumSize]]
    """

    def __init__(
        self,
        namespace: str,
        name: str,
        value: str,
        datum_type: str,
        max_size: int = 0,
        source_info: Optional["source_info"] = None,
    ) -> "pcd_typed":
        """Inits the Typed PCD Object."""
        super().__init__(namespace, name, value, source_info)
        self.datum_type = datum_type
        self.max_size = int(max_size)

    def __repr__(self) -> str:
        """A string representation of the object."""
        return f"{self.namespace}.{self.name} = {self.value} |{self.datum_type}|{self.max_size} @ {self.source_info}"


pcd_variable_attributes = ["NV", "BS", "RT", "RO"]


class pcd_variable(pcd):
    """A Variable PCD object.

    Attributes:
        var_name: VariableName
        var_guid: VariableGuid
        var_offset: VariableOffset
        default: default value
        attribute: Optional attributes
        source_info: source info

    EXAMPLE: PcdTokenSpaceGuidCName.PcdCName|VariableName|VariableGuid|VariableOffset[|HiiDefaultValue[|HiiAttribute]]
    """

    def __init__(
        self,
        namespace: str,
        name: str,
        var_name: str,
        var_guid: str,
        var_offset: str,
        default: Optional[str] = None,
        attributes: Optional[str] = None,
        source_info: Optional["source_info"] = None,
    ) -> "pcd_variable":
        """Inits a pcd_variable object."""
        super().__init__(namespace, name, "", source_info)
        if attributes is None:
            attributes = []
        self.var_name = var_name
        self.var_guid = var_guid
        self.var_offset = var_offset
        self.default = default

        if isinstance(attributes, str):
            attributes = attributes.split(",")
        attributes = [str(x).upper().strip() for x in attributes]
        if any([x not in pcd_variable_attributes for x in attributes]):
            raise ValueError(f"Invalid PcdHiiAttribute values: {attributes}")
        self.attributes = attributes

    def __repr__(self) -> str:
        """A string representation of the object."""
        pcd_data = f"{self.var_guid}|{self.var_offset}|{self.default}|{self.attributes}"
        return f"{self.namespace}.{self.name} = {self.var_name} |{pcd_data} @ {self.source_info}"


class build_option:
    """Build option object.

    Attributes:
        tool_code: One of the defined tool codes in the Conf/tools_def.txt file.
        attribute: for example flags, d_path, path
        data: the actual flags or path you want to set
        target: DEBUG, RELEASE, or other
        tagname: the tool chain tag
        arch: ARM, AARCH64, IA32, X64, etc
        family: Conf/tools_def.txt defines FAMILY as one of MSFT, INTEL or GCC.
        replace: whether or not this replaces the default from tools_def, if this is false, we append
        source_info: source_info object.

    NOTE: Contains the data for a build option
        EX: MSFT:*_*_*_CC_FLAGS = /D MDEPKG_NDEBUG
    """

    # {FAMILY}:{TARGET}_{TAGNAME}_{ARCH}_{TOOLCODE}_{ATTRIBUTE}
    def __init__(
        self,
        tool_code: str,
        attribute: str,
        data: str,
        target: str = "*",
        tagname: str = "*",
        arch: str = "*",
        family: Optional[str] = None,
        replace: bool = False,
        source_info: "source_info" = None,
    ) -> "build_option":
        """Inits a build_option object.

        Args:
            tool_code: One of the defined tool codes in the Conf/tools_def.txt file.
            attribute: for example flags, d_path, path
            data: the actual flags or path you want to set
            target: DEBUG, RELEASE, or other
            tagname: the tool chain tag
            arch: ARM, AARCH64, IA32, X64, etc
            family: Conf/tools_def.txt defines FAMILY as one of MSFT, INTEL or GCC.
            replace: whether or not this replaces the default from tools_def, if this is false, we append
            source_info: source_info object.

        NOTE: attribute Attribute:
            The flags defined in this section are appended to flags defined in the tools_def.txt file
            for individual tools.

        NOTE: fam9ily Attribute:
            Typically, this field is used to help the build tools determine whether the line
            is used for Microsoft style Makefiles or the GNU style Makefile
        """
        self.family = family
        self.target = target
        self.tagname = tagname
        self.arch = arch
        self.tool_code = tool_code
        self.attribute = attribute
        self.replace = replace
        self.data = data
        self.source_info = source_info

    def __eq__(self, other: "build_option") -> bool:
        """Enables equality comparisons (a == b)."""
        if type(other) is not build_option:
            return False
        if self.family != other.family:
            return False
        if self.target != other.target:
            return False
        if self.tagname != other.tagname:
            return False
        if self.arch != other.arch:
            return False
        if self.tool_code != other.tool_code:
            return False
        if self.attribute != other.attribute:
            return False
        return True

    def __hash__(self) -> int:
        """Returns the hash of an object for hashtables."""
        return hash(self.__repr__(False))

    def __repr__(self, include_data: bool = True) -> str:
        """A string representation of the object."""
        rep = "" if self.family is None else f"{self.family}:"
        rep += "_".join((self.target, self.tagname, self.arch, self.tool_code, self.attribute))
        if include_data:
            rep += f"= {self.data}"
        return rep


class default_store:
    """Object containing information on a default store.

    Attributes:
        index: index
        value: defaults to "Standard"
        source_info: defaults to None
    """

    """ contains the information on a default store. """
    """ 0 | Standard        # UEFI Standard default """

    def __init__(self, index: int = 0, value: str = "Standard", source_info: Optional[str] = None) -> "default_store":
        """Inits a default store object.

        NOTE: Local means DEFINE is in front and is localized to that particular section
        """
        self.index = int(index)
        self.value = value
        self.source_info = source_info

    def __repr__(self) -> str:
        """A string representation of the object."""
        return f"{self.index} | {self.value}"

    def __hash__(self) -> int:
        """Returns the hash of an object for hashtables."""
        return hash(self.index)

    def __eq__(self, other: "default_store") -> bool:
        """Enables equality comparisons (a == b)."""
        if type(other) is not default_store:
            return False
        return other.index == self.index


class source_info:
    """Object representing source file information.

    Attributes:
        file (str): filename
        lineno (:obj:`int`, optional): line number
    """

    def __init__(self, file: str, lineno: int = None) -> "source_info":
        """Inits a source_info object."""
        self.file = file
        self.lineno = lineno

    def __repr__(self) -> str:
        """Returns a string representation of object."""
        if self.lineno is None:
            return self.file
        return f"{self.file}:{self.lineno}"



# @file dsc_parser.py
# Code to help parse DSC files
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code to help parse DSC files."""

import logging
import os
import re
from typing import List, Optional

from edk2toollib.uefi.edk2.parsers.base_parser import HashFileParser


class DscParser(HashFileParser):
    """Object representing a parsed DSC file with a capability to parse.

    Attributes:
        SixMods (list): list of X64 Modules (the line in the file)
        SixModsEnhanced (list): Better parsed (in a dict) list of X64 Modules
        ThreeMods (list): list of IA32 Modules (the line in the file)
        ThreeModsEnhanced (list): Better parsed (in a dict) list of IA32 Modules
        OtherMods (list): list of other Mods that are not IA32, X64 specific
        Libs (list): list of Libs (the line in the file)
        LibsEnhanced (list): Better parsed (in a dict) list of Libs
        ScopedLibraryDict (dict): key (library class) value (list of instances)
        LibraryClassToInstanceDict (dict): Key (Library class) Value (Instance)
        Pcds (list): List of Pcds

    !!! note
        ScopedLibraryDict can have multiple library instances for the same scope because the INF
        can also filter the module types it supports. For example, two library instances could be
        in the scope of "common", but one filters to only PEI (MyLib| PEI_CORE) and the other
        filters to PEIM (MyLib| PEIM).
    """

    SECTION_LIBRARY = "libraryclasses"
    SECTION_COMPONENT = "components"
    SECTION_REGEX = re.compile(r"\[(.*)\]")
    OVERRIDE_REGEX = re.compile(r"\<(.*)\>")

    def __init__(self) -> "HashFileParser":
        """Init an empty Parser."""
        super(DscParser, self).__init__("DscParser")
        self.SixMods = []
        self.SixModsEnhanced = []
        self.ThreeMods = []
        self.ThreeModsEnhanced = []
        self.OtherMods = []
        self.Libs = []
        self.Components = []
        self.LibsEnhanced = []
        self.ScopedLibraryDict = {}
        self.ParsingInBuildOption = 0
        self.LibraryClassToInstanceDict = {}
        self.Pcds = []
        self.PcdValueDict = {}
        self._no_fail_mode = False
        self._dsc_file_paths = set()  # This includes the full paths for every DSC that makes up the file

    def ReplacePcds(self, line: str) -> str:
        """Attempts to replace a token if it is a PCD token."""
        if line.startswith("!if"):
            tokens = line.split()
            if tokens[1] in self.PcdValueDict:
                line = line.replace(tokens[1], self.PcdValueDict[tokens[1]])
        return line

    def __ParseLine(self, Line: str, file_name: Optional[str] = None, lineno: int = None) -> tuple:
        line_stripped = self.StripComment(Line).strip()
        if len(line_stripped) < 1:
            return ("", [], None)
        line_stripped = self.ReplacePcds(line_stripped)
        line_resolved = self.ReplaceVariables(line_stripped)
        if self.ProcessConditional(line_resolved):
            # was a conditional
            # Other parser returns line_resolved, [].  Need to figure out which is right
            return ("", [], None)

        # not conditional keep processing

        # check if conditional is active
        if not self.InActiveCode():
            return ("", [], None)

        # check for include file and import lines from file
        if line_resolved.strip().lower().startswith("!include"):
            # include line.
            tokens = line_resolved.split()
            include_file = tokens[1]
            sp = self.FindPath(include_file)
            if sp is None:
                raise FileNotFoundError(include_file)
            self.Logger.debug("Opening Include File %s" % sp)
            self._PushTargetFile(sp)
            lf = open(sp, "r")
            loc = lf.readlines()
            lf.close()
            return ("", loc, sp)

        # check for new section
        (IsNew, Section) = self.ParseNewSection(line_resolved)
        if IsNew:
            self.CurrentSection = Section.upper()
            self.Logger.debug("New Section: %s" % self.CurrentSection)
            self.Logger.debug("FullSection: %s" % self.CurrentFullSection)
            return (line_resolved, [], None)

        # process line in x64 components
        if self.CurrentFullSection.upper() == "COMPONENTS.X64":
            if self.ParsingInBuildOption > 0:
                if ".inf" in line_resolved.lower():
                    p = self.ParseInfPathLib(line_resolved)
                    self.Libs.append(p)
                    self.Logger.debug("Found Library in a 64bit BuildOptions Section: %s" % p)
                elif self.RegisterPcds(line_resolved):
                    self.Logger.debug("Found a Pcd in a 64bit Module Override section")
            else:
                if ".inf" in line_resolved.lower():
                    p = self.ParseInfPathMod(line_resolved)
                    self.SixMods.append(p)
                    if file_name is not None and lineno is not None:
                        self.SixModsEnhanced.append({"file": os.path.normpath(file_name), "lineno": lineno, "data": p})
                    self.Logger.debug("Found 64bit Module: %s" % p)

            self.ParsingInBuildOption = self.ParsingInBuildOption + line_resolved.count("{")
            self.ParsingInBuildOption = self.ParsingInBuildOption - line_resolved.count("}")
            return (line_resolved, [], None)

        # process line in ia32 components
        elif self.CurrentFullSection.upper() == "COMPONENTS.IA32":
            if self.ParsingInBuildOption > 0:
                if ".inf" in line_resolved.lower():
                    p = self.ParseInfPathLib(line_resolved)
                    self.Libs.append(p)
                    if file_name is not None and lineno is not None:
                        self.LibsEnhanced.append({"file": os.path.normpath(file_name), "lineno": lineno, "data": p})
                    self.Logger.debug("Found Library in a 32bit BuildOptions Section: %s" % p)
                elif self.RegisterPcds(line_resolved):
                    self.Logger.debug("Found a Pcd in a 32bit Module Override section")

            else:
                if ".inf" in line_resolved.lower():
                    p = self.ParseInfPathMod(line_resolved)
                    self.ThreeMods.append(p)
                    if file_name is not None and lineno is not None:
                        self.ThreeModsEnhanced.append(
                            {"file": os.path.normpath(file_name), "lineno": lineno, "data": p}
                        )
                    self.Logger.debug("Found 32bit Module: %s" % p)

            self.ParsingInBuildOption = self.ParsingInBuildOption + line_resolved.count("{")
            self.ParsingInBuildOption = self.ParsingInBuildOption - line_resolved.count("}")
            return (line_resolved, [], None)

        # process line in other components
        elif "COMPONENTS" in self.CurrentFullSection.upper():
            if self.ParsingInBuildOption > 0:
                if ".inf" in line_resolved.lower():
                    p = self.ParseInfPathLib(line_resolved)
                    self.Libs.append(p)
                    self.Logger.debug("Found Library in a BuildOptions Section: %s" % p)
                elif self.RegisterPcds(line_resolved):
                    self.Logger.debug("Found a Pcd in a Module Override section")

            else:
                if ".inf" in line_resolved.lower():
                    p = self.ParseInfPathMod(line_resolved)
                    self.OtherMods.append(p)
                    self.Logger.debug("Found Module: %s" % p)

            self.ParsingInBuildOption = self.ParsingInBuildOption + line_resolved.count("{")
            self.ParsingInBuildOption = self.ParsingInBuildOption - line_resolved.count("}")
            return (line_resolved, [], None)

        # process line in library class section (don't use full name)
        elif self.CurrentSection.upper() == "LIBRARYCLASSES":
            if ".inf" in line_resolved.lower():
                p = self.ParseInfPathLib(line_resolved)
                self.Libs.append(p)
                self.Logger.debug("Found Library in Library Class Section: %s" % p)
            return (line_resolved, [], None)
        # process line in PCD section
        elif self.CurrentSection.upper().startswith("PCDS"):
            if self.RegisterPcds(line_resolved):
                self.Logger.debug("Found a Pcd in a PCD section")
            return (line_resolved, [], None)
        else:
            return (line_resolved, [], None)

    def __ParseDefineLine(self, Line: str) -> tuple:
        line_stripped = self.StripComment(Line).strip()
        if len(line_stripped) < 1:
            return ("", [])

        # this line needs to be here to resolve any symbols inside the !include lines, if any
        self.RegisterPcds(line_stripped)
        line_stripped = self.ReplacePcds(line_stripped)
        line_resolved = self.ReplaceVariables(line_stripped)
        if self.ProcessConditional(line_resolved):
            # was a conditional
            # Other parser returns line_resolved, [].  Need to figure out which is right
            return ("", [])

        # not conditional keep processing

        # check if conditional is active
        if not self.InActiveCode():
            return ("", [])

        # check for include file and import lines from file
        if line_resolved.strip().lower().startswith("!include"):
            # include line.
            tokens = line_resolved.split()
            include_file = tokens[1]
            self.Logger.debug("Opening Include File %s" % include_file)
            sp = self.FindPath(include_file)
            if sp is None:
                raise FileNotFoundError(include_file)
            self._PushTargetFile(sp)
            lf = open(sp, "r")
            loc = lf.readlines()
            lf.close()
            return ("", loc)

        # check for new section
        (IsNew, Section) = self.ParseNewSection(line_resolved)
        if IsNew:
            self.CurrentSection = Section.upper()
            self.Logger.debug("New Section: %s" % self.CurrentSection)
            self.Logger.debug("FullSection: %s" % self.CurrentFullSection)
            return (line_resolved, [])

        # process line based on section we are in
        if (self.CurrentSection == "DEFINES") or (self.CurrentSection == "BUILDOPTIONS"):
            if line_resolved.count("=") >= 1:
                tokens = line_resolved.split("=", 1)
                leftside = tokens[0].split()
                if len(leftside) == 2:
                    left = leftside[1]
                else:
                    left = leftside[0]
                right = tokens[1].strip()

                self.LocalVars[left] = right
                self.Logger.debug("Key,values found:  %s = %s" % (left, right))

                # iterate through the existed LocalVars and try to resolve the symbols
                for var in self.LocalVars:
                    self.LocalVars[var] = self.ReplaceVariables(self.LocalVars[var])
                return (line_resolved, [])
        else:
            return (line_resolved, [])

    def ParseInfPathLib(self, line: str) -> str:
        """Parses a line with an INF path Lib."""
        if line.count("|") > 0:
            line_parts = []
            c = line.split("|")[0].strip()
            i = line.split("|")[1].strip()
            if c in self.LibraryClassToInstanceDict:
                line_parts = self.LibraryClassToInstanceDict.get(c)
            sp = self.FindPath(i)
            line_parts.append(sp)
            self.LibraryClassToInstanceDict[c] = line_parts
            return line.split("|")[1].strip()
        else:
            return line.strip().split()[0]

    def ParseInfPathMod(self, line: str) -> str:
        """Parses a line with an INF path."""
        return line.strip().split()[0].rstrip("{")

    def __ProcessMore(self, lines: list, file_name: Optional[str] = None) -> None:
        """Runs after ProcessDefines and does a full parsing of the DSC.

        Everything is resolved to a final state
        """
        if len(lines) == 0:
            return
        for index in range(len(lines)):
            # we try here so that we can catch exceptions from individual lines
            try:
                raw_line = lines[index]
                (line, add, new_file) = self.__ParseLine(raw_line, file_name=file_name, lineno=index + 1)
                if len(line) > 0:
                    self.Lines.append(line)
                self.__ProcessMore(add, file_name=new_file)
            except Exception as e:
                # check if we're in no fail mode or not
                if not self._no_fail_mode:  # if we are, fail
                    raise
                else:
                    # otherwise, let the user know that we failed in the DSC
                    self.Logger.warning(f"DSC Parser (No-Fail Mode): {raw_line}")
                    self.Logger.warning(e)

    def __ProcessDefines(self, lines: list) -> None:
        """Goes through a file once to look for [Define] sections.

        Only Sections, DEFINE, X = Y, and !includes are resolved
        This resolves all the defines since they can be anywhere in a file.
        Ideally this should be run until we reach stable state but this parser is not
        accurate and is more of an approximation of what the real parser does.
        """
        if len(lines) == 0:
            return
        for raw_line in lines:
            # we want to catch exceptions here since we are doing includes as we potentially might blow up
            # we want to catch on a line by line basis
            try:
                (line, add) = self.__ParseDefineLine(raw_line)
                self.__ProcessDefines(add)
            except Exception:
                # Since we're going to do this in ProcessMore, don't warn people if there's an exception
                # otherwise, raise the exception and act normally
                if not self._no_fail_mode:
                    raise
        # Reset the PcdValueDict as this was just to find any Defines.
        self.PcdValueDict = {}

    def _parse_libraries(self) -> None:
        """Builds a lookup table of all possible library instances depending on scope.

        The following is the key/value pair:
        key: The library class name with the scope appended. Examples below:
            $(LIB_NAME).$(ARCH).$(MODULE_TYPE)
            $(LIB_NAME).common.$(MODULE_TYPE)
            $(LIB_NAME).$(ARCH)
            $(LIB_NAME).common
        """
        current_scopes = []
        for line in self.Lines:
            current_scopes = self._get_current_scope(current_scopes, line.lower(), self.SECTION_LIBRARY.lower())

            # The current section is not SECTION_LIBRARY, so we have no valid scopes. continue to next line.
            if not current_scopes:
                continue

            # This line is starting a new section with a new scope. Start reading the new line
            if self.SECTION_REGEX.match(line):
                continue

            if len(line.split("|")) != 2:
                logging.debug("Unexpected Line in Library Section:")
                logging.debug(f"  {line}")
                continue

            # We are in a valid section, so lets parse the line and add it to our dictionary.
            lib, instance = tuple(line.split("|"))
            for scope in current_scopes:
                key = f"{scope.strip()}.{lib.strip()}".lower()
                value = instance.strip()
                if os.path.isabs(value):
                    value = self._Edk2PathUtil.GetEdk2RelativePathFromAbsolutePath(value)
                if key in self.ScopedLibraryDict and value not in self.ScopedLibraryDict[key]:
                    self.ScopedLibraryDict[key].insert(0, value)
                else:
                    self.ScopedLibraryDict[key] = [value]

        return

    def _parse_components(self) -> None:
        current_scopes = []
        lines = iter(self.Lines)

        try:
            while True:
                line = next(lines)

                current_scopes = self._get_current_scope(current_scopes, line.lower(), self.SECTION_COMPONENT)
                library_override_dict = {"NULL": []}

                # The current section is not SECTION_COMPONENT, so we have no valid scopes. continue to next line.
                if not current_scopes:
                    continue

                # This line is starting a new section with a new scope. Start reading the new line
                if self.SECTION_REGEX.match(line.lower()):
                    continue

                # This component has overrides we need to handle
                if line.strip().endswith("{"):
                    line = str(line)
                    logging.debug(f"Building Library Override Dictionary for Component: {line.strip(' {')}")
                    library_override_dict = self._build_library_override_dictionary(lines)

                for scope in current_scopes:
                    # Components without a specific scope (common or empty) are added to all current scopes
                    if os.path.isabs(line.strip(" {")):
                        line = self._Edk2PathUtil.GetEdk2RelativePathFromAbsolutePath(line.strip(" {"))
                    if scope == "common":
                        for arch in self.LocalVars["SUPPORTED_ARCHITECTURES"].split("|"):
                            self.Components.append((line.strip(" {"), arch, library_override_dict))
                    else:
                        self.Components.append((line.strip(" {"), scope, library_override_dict))

        except StopIteration:
            return

    def _get_current_scope(self, scope_list: list[str], line: str, section_type: str) -> list[str]:
        """Returns the list of scopes that this line is in, as long as the section_type is correct.

        Scopes can be different depending on the section type. Component sections can only
        contain a single scope, but library sections can contain multiple scopes.

        !!! warning
            The returned list of scopes does not include the section type.
        """
        match = self.SECTION_REGEX.match(line)

        # If the line is not a section header, return the old section
        if not match:
            return scope_list

        # If the line is a section header, but not the correct section type, return []
        elif not match.group().startswith(f"[{section_type}"):
            return []

        # The line must be a section header and of the correct section type. Return it
        current_section = []
        section_list = match.group().strip("[]").split(",")

        for section in section_list:
            # Remove the section type and strip the leftover '.'. If it's empty after that, then it is actually "common"
            section = section.replace(section_type, "").replace("Common", "common").strip().lstrip(".")
            current_section.append(section or "common")
        return current_section

    def _build_library_override_dictionary(self, lines: List[str]) -> None:
        library_override_dictionary = {"NULL": []}
        section = ""

        for line in lines:
            l_line = line.lower().strip()

            if l_line == "}":
                break

            if self.OVERRIDE_REGEX.match(l_line):
                if l_line == f"<{self.SECTION_LIBRARY}>":
                    section = self.SECTION_LIBRARY
                else:  # Add more sections here if needed
                    section = ""
                continue

            if section == self.SECTION_LIBRARY:
                logging.debug(f"  Library Section Override: {line}")
                lib, instance = map(str.strip, line.split("|"))

                if os.path.isabs(instance):
                    instance = self._Edk2PathUtil.GetEdk2RelativePathFromAbsolutePath(instance)

                lib = lib.lower()
                if lib == "null":
                    library_override_dictionary["NULL"].append(instance)
                else:
                    library_override_dictionary[lib] = instance
        return library_override_dictionary

    def SetNoFailMode(self, enabled: bool = True) -> None:
        """The parser won't throw exceptions when this is turned on.

        WARNING: This can result in some weird behavior
        """
        self._no_fail_mode = enabled

    def ResetParserState(self) -> None:
        """Resets the parser."""
        #
        # add more DSC parser based state reset here, if necessary
        #
        super(DscParser, self).ResetParserState()

    def ParseFile(self, filepath: str) -> None:
        """Parses the DSC file at the provided path."""
        self.Logger.debug("Parsing file: %s" % filepath)
        sp = self.FindPath(filepath)
        if sp is None:
            raise FileNotFoundError(filepath)
        self._PushTargetFile(sp)
        f = open(sp, "r")
        # expand all the lines and include other files
        file_lines = f.readlines()
        self.__ProcessDefines(file_lines)
        # reset the parser state before processing more
        self.ResetParserState()
        self._PushTargetFile(sp)
        self.__ProcessMore(file_lines, file_name=sp)
        f.close()

        self._parse_libraries()
        self._parse_components()
        self.Parsed = True

    def _PushTargetFile(self, targetFile: str) -> None:
        self.TargetFilePath = os.path.abspath(targetFile)
        self._dsc_file_paths.add(self.TargetFilePath)

    def GetMods(self) -> list:
        """Returns a list with all Mods."""
        return self.ThreeMods + self.SixMods

    def GetModsEnhanced(self) -> list:
        """Returns a list with all ModsEnhanced."""
        return self.ThreeModsEnhanced + self.SixModsEnhanced

    def GetLibs(self) -> list:
        """Returns a list with all Libs."""
        return self.Libs

    def GetLibsEnhanced(self) -> list:
        """Returns a list with all LibsEnhanced."""
        return self.LibsEnhanced

    def GetAllDscPaths(self) -> set:
        """Returns a list with all the paths that this DSC uses (the base file and any includes).

        They are not all guaranteed to be DSC files
        """
        return self._dsc_file_paths

    def RegisterPcds(self, line: str) -> None:
        """Reads the line and registers any PCDs found."""
        if "tokenspaceguid" in line.lower() and line.count("|") > 0 and line.count(".") > 0:
            # should be a pcd statement
            p = line.partition("|")
            self.Pcds.append(p[0].strip())
            self.PcdValueDict[p[0].strip()] = p[2].strip()
            self.Logger.debug("Found a Pcd: %s" % p[0].strip())
            return True
        return False



# @file locate_tools.py
# This module provides python services that locate common development tools using vswhere.exe,
# vsvars.bat, and other Windows based tools.  This works best on systems running Windows
# and with dev tools installed but will attempt to use known paths for WinSDK if the dev
# tools are not available.  This is only a best effort to locate the SDK tools in well
# known/default install locations.
#
# Suggested Dev Tools:
#   Current Windows SDKs
#   Visual Studio 2017 Build Tools or newer
#
# Note: When this module is used in local develop mode it will download vswhere.exe from the github.
#       It will confirm the hash is a known good value.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module to assist in locating tools via VsWhere.

NOTE: Has the capability to download VSwhere.
"""

import glob
import importlib.resources as resources
import logging
import os
import re
import subprocess
from pathlib import Path
from typing import List, Optional

from edk2toollib.utility_functions import GetHostInfo, RunCmd

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import urllib.error
import urllib.request

# Update this when you want a new version of VsWhere
__VERSION = "3.1.7"
__URL = "https://github.com/microsoft/vswhere/releases/download/{}/vswhere.exe".format(__VERSION)
__SHA256 = "c54f3b7c9164ea9a0db8641e81ecdda80c2664ef5a47c4191406f848cc07c662"

#
# Supported Versions that can be queried with vswhere
# Use lower case for key as all comparisons will be lower case
#
supported_vs_versions = {"vs2017": "15.0,16.0", "vs2019": "16.0,17.0", "vs2022": "17.0,18.0"}


# Downloads VSWhere
def _DownloadVsWhere(unpack_folder: os.PathLike = None) -> None:
    if unpack_folder is None:
        unpack_folder = os.path.dirname(__VsWherePath())

    out_file_name = os.path.join(unpack_folder, "vswhere.exe")
    logging.info("Attempting to download vswhere to: {}. This may take a second.".format(unpack_folder))

    # check if the path we want to download to exists
    if not os.path.exists(unpack_folder):
        os.makedirs(unpack_folder, exist_ok=True)
        logging.info("Created folder: {}".format(unpack_folder))

    # check if we have the vswhere file already downloaded
    if not os.path.isfile(out_file_name):
        try:
            # Download the file and save it locally under `temp_file_name`
            with urllib.request.urlopen(__URL) as response, open(out_file_name, "wb") as out_file:
                out_file.write(response.read())
        except urllib.error.HTTPError as e:
            logging.error("We ran into an issue when getting VsWhere")
            raise e

    # do the hash to make sure the file is good
    with open(out_file_name, "rb") as file:
        import hashlib

        temp_file_sha256 = hashlib.sha256(file.read()).hexdigest()
    if temp_file_sha256 != __SHA256:
        # delete the file since it's not what we're expecting
        os.remove(out_file_name)
        raise RuntimeError(f"VsWhere - sha256 does not match\n\tdownloaded:\t{temp_file_sha256}\n\t")


def __VsWherePath() -> str:
    file = "vswhere.exe"
    return str(resources.files("edk2toollib") / "bin" / file)


def GetVsWherePath(fail_on_not_found: bool = True) -> Optional[str]:
    """Get the VsWhere tool path.

    Args:
        fail_on_not_found (:obj:`bool`, optional): whether to log an error or not.

    Returns:
        (str): "/PATH/TO/vswhere.exe" if found
        (None): if fail_on_not_found and is not found
    """
    if GetHostInfo().os != "Windows":
        raise EnvironmentError("VsWhere cannot be used on a non-windows system.")
    vswhere_path = __VsWherePath()
    # check if we can't find it, look for vswhere in the path
    if not os.path.isfile(vswhere_path):
        for env_var in os.getenv("PATH").split(os.pathsep):
            env_var = os.path.join(os.path.normpath(env_var), "vswhere.exe")
            if os.path.isfile(env_var):
                vswhere_path = env_var
                break

    # if we still can't find it, download it
    if not os.path.isfile(vswhere_path):
        vswhere_dir = os.path.dirname(vswhere_path)
        try:  # try to download
            _DownloadVsWhere(vswhere_dir)
        except Exception as exc:
            logging.warning("Tried to download VsWhere and failed with exception: %s", str(exc))

    # if we're still hosed
    if not os.path.isfile(vswhere_path) and fail_on_not_found:
        logging.error("We weren't able to find vswhere!")
        return None

    return vswhere_path


def FindWithVsWhere(products: str = "*", vs_version: Optional[str] = None) -> Optional[str]:
    """Finds a product with VS Where. Returns the newest match.

    Args:
        products (:obj:`str`, optional): product defined by vswhere tool
        vs_version (:obj:`str`, optional): helper to find version of supported VS version (example vs2019)

    Returns:
        (str): VsWhere products
        (None): If no products returned
    Raises:
        (EnvironmentError): Not on a windows system or cannot locate the VsWhere
        (ValueError): Unsupported VS version
        (RuntimeError): Error when running vswhere
    """
    results = FindAllWithVsWhere(products, vs_version)
    return results[0] if results else None


def FindAllWithVsWhere(products: str = "*", vs_version: Optional[str] = None) -> Optional[List[str]]:
    """Finds a product with VS Where. Returns all matches, sorted by newest version.

    Args:
        products (:obj:`str`, optional): product defined by vswhere tool
        vs_version (:obj:`str`, optional): helper to find version of supported VS version (example vs2019)

    Returns:
        (list[str]): VsWhere products
        (None): If no products returned
    Raises:
        (EnvironmentError): Not on a windows system or cannot locate the VsWhere
        (ValueError): Unsupported VS version
        (RuntimeError): Error when running vswhere
    """
    cmd = "-nologo -all -sort -property installationPath"
    vs_where_path = GetVsWherePath()  # Will raise the Environment Error if not on windows.
    if vs_where_path is None:
        raise EnvironmentError("Unable to locate the VsWhere Executable.")

    if products is not None:
        cmd += " -products " + products
    if vs_version is not None:
        vs_version = vs_version.lower()
        if vs_version in supported_vs_versions.keys():
            cmd += " -version " + supported_vs_versions[vs_version]
        else:
            raise ValueError(f"{vs_version} unsupported. Supported Versions: {' '.join(supported_vs_versions)}")
    a = StringIO()
    ret = RunCmd(vs_where_path, cmd, outstream=a)
    if ret != 0:
        a.close()
        raise RuntimeError(f"Unkown Error while executing VsWhere: errcode {ret}.")
    p1 = a.getvalue().strip()
    a.close()
    if len(p1.strip()) > 0:
        return [line.strip() for line in p1.splitlines()]
    return None


def QueryVcVariables(
    keys: list, arch: str = None, product: str = None, vs_version: str = None, vc_version: str = None
) -> str:
    """Launch vcvarsall.bat and read the settings from its environment.

    This is a windows only function and Windows is case insensitive for the keys

    Args:
        keys (list): names of env variables to collect after bat run
        arch (:obj:`str`, optional): arch to run
        product (:obj:`str`, optional): values defined by vswhere.exe
        vs_version (:obj:`str`, optional): helper to find version of supported VS version (example vs2019)
        vc_version (:obj:`str`, optional): Exact compiler version to consider for variables (example 14.42.34433)

    Returns:
        (dict): the appropriate environment variables
    """
    if product is None:
        product = "*"
    if arch is None:
        # TODO: look up host architecture?
        arch = "amd64"
    interesting = set(x.upper() for x in keys)
    result = {}

    # Handle failing to get the vs_path from FindWithVsWhere
    try:
        vs_path_list = FindAllWithVsWhere(product, vs_version)
    except (EnvironmentError, ValueError, RuntimeError) as e:
        logging.error(str(e))
        raise
    if vs_path_list is None:
        err_msg = "VS path not found."
        if vs_version is not None:
            err_msg += f" Might need to verify {vs_version} install exists."
        else:
            err_msg += " Might need to specify VS version... i.e. vs2022."
        logging.error(err_msg)
        raise ValueError(err_msg)

    if len(os.environ["PATH"]) > 8191:
        w = (
            "Win32 Command prompt ignores any environment variables longer then 8191 characters, but your path is "
            f"{len(os.environ['PATH'])} characters. Due to this, PATH will not be used when searching for "
            f"[{interesting}]. This could result in missing keys."
        )
        logging.warning(w)

    # Attempt to find vcvarsall.bat from any of the found VS installations
    vcvarsall_path = next(
        (
            Path(vs_path, "VC", "Auxiliary", "Build", "vcvarsall.bat")
            for vs_path in vs_path_list
            if Path(vs_path, "VC", "Auxiliary", "Build", "vcvarsall.bat").exists()
        ),
        None,
    )

    if vcvarsall_path is None:
        e = f"Could not locate VC/Auxiliary/Build/vcvarsall.bat in [{vs_path_list}]"
        logging.error(e)
        raise ValueError(e)

    if vc_version:
        vc_version = f"-vcvars_ver={vc_version}"
    else:
        vc_version = ""

    logging.debug("Calling '%s %s %s'", vcvarsall_path, arch, vc_version)
    popen = subprocess.Popen(
        '"%s" %s %s & set' % (vcvarsall_path, arch, vc_version), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    try:
        stdout, stderr = popen.communicate()
        if popen.wait() != 0:
            stderr = stderr.decode("mbcs")
            if stderr.startswith("The input line is too long"):
                stderr = (
                    ".bat cmd can not handle args greater than 8191 chars; your total ENV var len exceeds 8191. "
                    "Reduce the total length of your ENV variables to resolve this (Typically your PATH is "
                    "too long)."
                )
            logging.error(stderr)
            raise RuntimeError(stderr)
        stdout = stdout.decode("mbcs")
        for line in stdout.split("\n"):
            if "=" not in line:
                continue
            line = line.strip()
            key, value = line.split("=", 1)
            if key.upper() in interesting:
                if value.endswith(os.pathsep):
                    value = value[:-1]
                result[key] = value
    finally:
        popen.stdout.close()
        popen.stderr.close()

    if len(result) != len(interesting):
        logging.debug("Input: " + str(sorted(interesting)))
        logging.debug("Result: " + str(sorted(list(result.keys()))))
        result_set = set([key.upper() for key in result.keys()])
        difference = list(interesting.difference(result_set))

        logging.error("We were not able to find on the keys requested from vcvarsall.")
        logging.error("We didn't find: %s" % str(difference))
        raise ValueError("Missing keys when querying vcvarsall: " + str(difference))
    return result


def _CompareWindowVersions(a: str, b: str) -> int:
    """Compares Windows versions.

    Returns:
        (int): 1 if a > b
        (int): 0 if a == b
        (int): -1 if a < b
    """
    a_periods = str(a).count(".")
    b_periods = str(b).count(".")
    if a_periods == 3 and b_periods != 3:
        return 1
    if b_periods == 3 and a_periods != 3:
        return -1
    if a_periods != 3 and b_periods != 3:
        return 0
    a_parts = str(a).split(".")
    b_parts = str(b).split(".")
    for i in range(3):
        a_p = int(a_parts[i])
        b_p = int(b_parts[i])
        if a_p > b_p:
            return 1
        if a_p < b_p:
            return -1
    return 0


def _CheckArchOfMatch(match: str) -> bool:
    """Returns if this binary matches our host.

    Returns:
        (bool): if arch matches

    NOTE: if no arch is in the match, then we return true
    """
    match = str(match).lower()
    isx86 = "x86" in match
    isx64 = "x64" in match or "amd64" in match
    isArm64 = "aarch" in match or "aarch64" in match or "arm64" in match
    isi386 = "i386" in match
    isArm = not isArm64 and ("arm" in match)
    count = 0
    count += 1 if isx64 else 0
    count += 1 if isx86 else 0
    count += 1 if isArm else 0
    count += 1 if isArm64 else 0
    count += 1 if isi386 else 0
    if count == 0:  # we don't know what arch this is?
        return True
    if count > 1:  # there are more than one arch for this binary
        logging.warning("We found more than one architecture for {}. Results maybe inconsistent".format(match))
        return True

    _, arch, bits = GetHostInfo()
    bits = int(bits)
    if isx86 and (bits < 32 or arch != "x86"):
        return False
    if isx64 and (bits < 64 or arch != "x86"):
        return False
    if isi386:
        # TODO add i386 to GetHostInfo
        return False
    if isArm64 and (bits < 64 or arch != "ARM"):
        return False
    if isArm and (bits < 32 or arch != "ARM"):
        return False
    return True


# does a glob in the folder that your sdk is
# uses the environmental variable WindowsSdkDir and tries to use WindowsSDKVersion
def FindToolInWinSdk(tool: str, product: str = None, arch: str = None) -> str:
    """Searches for a tool in the Windows SDK Directory.

    Args:
        tool (str): Tool to search for
        product (:obj:`str`, optional): product for searching with vswhere
        arch (:obj:`str`, optional): arch for searching with vswhere
    """
    variables = ["WindowsSdkDir", "WindowsSDKVersion"]
    # get the value with QueryVcVariables
    try:
        results = QueryVcVariables(variables, product, arch)
        # Get the variables we care about
        sdk_dir = results["WindowsSdkDir"]
        sdk_ver = results["WindowsSDKVersion"]
    except ValueError:
        sdk_dir = os.path.join(os.getenv("ProgramFiles(x86)"), "Windows Kits", "10", "bin")
        sdk_ver = "0.0.0.0"
    sdk_dir = os.path.realpath(sdk_dir)
    search_pattern = os.path.join(sdk_dir, "**", tool)

    match_offset = len(sdk_dir)
    # look for something like 10.0.12323.0123
    windows_ver_regex = re.compile(r"\d+\.\d+\.\d+\.\d+")
    top_version_so_far = -1
    top_match = None
    # Look at in match in the tree
    for match in glob.iglob(search_pattern, recursive=True):
        match_file = match[match_offset:]  # strip off the root
        match_front, match_end = os.path.split(match_file)  # split off the filename
        versions = windows_ver_regex.findall(match_front)  # look for windows versions
        top_ver_match = 0
        if not _CheckArchOfMatch(match_front):  # make sure it's a good arch for us
            continue
        if len(versions) == 0:
            top_ver_match = "0.0.0.0"  # if we have a bad version, we should fall to a bad version
        for version in versions:  # find the highest version if there are multiple in this?
            is_current_sdk_version = _CompareWindowVersions(version, sdk_ver) == 0
            if _CompareWindowVersions(version, top_ver_match) > 0 or is_current_sdk_version:
                top_ver_match = version
        # if we have a highest version or one that matches our current from environment variables?
        is_current_sdk_version = _CompareWindowVersions(top_ver_match, sdk_ver) == 0
        if _CompareWindowVersions(top_ver_match, top_version_so_far) > 0 or is_current_sdk_version:
            top_version_so_far = top_ver_match
            top_match = match
    if top_match is None:
        logging.critical("We weren't able to find {}".format(tool))
    return top_match



# @file tpm2_policy_calc.py
# This file contains classes used to calculate TPM 2.0 policies
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A module containing classes used to calculate TPM 2.0 policies."""

import hashlib
import struct
from typing import Optional, Union

import edk2toollib.tpm.tpm2_defs as t2d

# ========================================================================================
##
# POLICY TREE CLASSES
# These are used to describe a final policy structure.
# You can construct nodes to form complex policies from the policy primitive classes.
##
# PolicyTreeOr                                 <--- Tree Node
# /          \
# PolicyTreeSolo         PolicyTreeAnd                     <--- Tree Nodes
# /                     /        \
# PolicyCommandCode     PolicyLocality    PolicyCommandCode         <--- Primitives
##
# ========================================================================================


class PolicyHasher(object):
    """An object used to hash TPM 2.0 policies with a specified hash."""

    def __init__(self, hash_type: str) -> "PolicyHasher":
        """Init the hasher with the specified hash.

        Args:
            hash_type (str): sha256 or sha384

        Raises:
            (ValueError): Invalid hash type
        """
        if hash_type not in ["sha256", "sha384"]:
            raise ValueError("Invalid hash type '%s'!" % hash_type)

        self.hash_type = hash_type
        self.hash_size = {"sha256": 32, "sha384": 48}[hash_type]

    def get_size(self) -> int:
        """Returns the size of the hash."""
        return self.hash_size

    def hash(self, data: str) -> bytes:
        """Hashes the data using the specified type."""
        hash_obj = None
        if self.hash_type == "sha256":
            hash_obj = hashlib.sha256()
        else:
            hash_obj = hashlib.sha384()

        hash_obj.update(data)

        return hash_obj.digest()


class PolicyCalculator(object):  # noqa
    def __init__(self, primitive_dict, policy_tree):  # noqa
        # For now, we'll leave this pretty sparse.
        # We should have WAY more testing for this stuff.
        self.primitive_dict = primitive_dict
        self.policy_tree = policy_tree

    def generate_digest(self, digest_type):  # noqa
        pass


class PolicyTreeOr(object):
    """Object representing an OR junction in a policy tree."""

    def __init__(self, components: Union["PolicyTreeAnd", "PolicyTreeOr", "PolicyTreeSolo"]) -> "PolicyTreeOr":
        """Inits the policy tree junction with a list of connected components.

        Args:
            components (list): list of components

        Raises:
            (ValueError): More then 8 connections
        """
        # OR connections can only be 8 digests long.
        # They CAN, however, be links of ORs.
        if len(components) > 8:
            raise ValueError("OR junctions cannot contain more than 8 sub-policies!")

        self.components = components

    def get_type(self) -> str:
        """Returns the type of junction."""
        return "or"

    def validate(self) -> bool:
        """Validates all components have the correct attributes."""
        result = True

        for component in self.components:
            # All components must be convertible into a policy.
            if not hasattr(component, "get_policy"):
                result = False

            # All components must also be valid.
            if not hasattr(component, "validate") or not component.validate():
                result = False

        return result

    def get_policy_buffer(self, hash_obj: Union["PolicyTreeAnd", "PolicyTreeOr", "PolicyTreeSolo"]) -> bytes:
        """Creates and returns a buffer representing the policy."""
        concat_policy_buffer = b"\x00" * hash_obj.get_size()
        concat_policy_buffer += struct.pack(">L", t2d.TPM_CC_PolicyOR)
        concat_policy_buffer += b"".join([component.get_policy(hash_obj) for component in self.components])
        return concat_policy_buffer

    def get_policy(self, hash_obj: Union["PolicyTreeAnd", "PolicyTreeOr", "PolicyTreeSolo"]) -> int:
        """Returns a hashed policy buffer."""
        return hash_obj.hash(self.get_policy_buffer(hash_obj))


class PolicyTreeAnd(object):
    """Object representing an AND junction in a policy tree."""

    def __init__(self, components: list[Union["PolicyTreeAnd", "PolicyTreeOr", "PolicyTreeSolo"]]) -> "PolicyTreeAnd":
        """Inits the policy tree junction with a list of connected components."""
        # ANDs must only be composed of primitives. For simplicity, I guess.
        # Honestly, this has spiralled out of control, but something is better than nothing.
        for component in components:
            if not hasattr(component, "get_buffer_for_digest"):
                raise ValueError("AND junctions must consist of primitives!")

        self.components = components

    def get_type(self) -> str:
        """Returns the type of junction."""
        return "and"

    def validate(self) -> bool:
        """Validate."""
        # Not sure why don't validate here instead of in the init?
        return True

    def get_policy(self, hash_obj: Union["PolicyTreeAnd", "PolicyTreeOr", "PolicyTreeSolo"]) -> bytes:
        """Returns a hashed policy buffer."""
        current_digest = b"\x00" * hash_obj.get_size()
        for component in self.components:
            current_digest = hash_obj.hash(current_digest + component.get_buffer_for_digest())
        return current_digest


class PolicyTreeSolo(object):
    """This object should only be used to put a single policy claim under an OR."""

    def __init__(self, policy_obj: Union["PolicyTreeAnd", "PolicyTreeOr", "PolicyTreeSolo"]) -> "PolicyTreeSolo":
        """Inits the policy tree junction."""
        if not hasattr(policy_obj, "get_buffer_for_digest"):
            raise ValueError("Supplied policy object is missing required functionality!")

        self.policy_obj = policy_obj

    def get_type(self) -> str:
        """Returns the type of junction."""
        return "solo"

    def validate(self) -> bool:
        """Validate."""
        # Not sure why don't validate here instead of in the init?
        return True

    def get_policy_buffer(self, hash_obj: Union["PolicyTreeAnd", "PolicyTreeOr", "PolicyTreeSolo"]) -> bytes:
        """Creates and returns a buffer representing the policy."""
        return (b"\x00" * hash_obj.get_size()) + self.policy_obj.get_buffer_for_digest()

    def get_policy(self, hash_obj: Union["PolicyTreeAnd", "PolicyTreeOr"]) -> int:
        """Returns a hashed policy buffer."""
        return hash_obj.hash(self.get_policy_buffer(hash_obj))


# ========================================================================================
##
# POLICY PRIMITIVES
# These classes are used to describe a single assertion (eg. PolicyLocality) and
# can be used with the PolicyTree classes to construct complex policies.
##
# ========================================================================================


class PolicyLocality(object):
    """Policy Primitive to describe a single assertion to create complex assertions."""

    def __init__(self, localities: list[int]) -> "PolicyLocality":
        """Init with the requested localities."""
        # Update the bitfield with the requested localities.
        if localities is not None:
            self.bitfield = self.calc_bitfield_from_list(localities)
        else:
            self.bitfield = 0b00000000

    def get_bitfield(self) -> int:
        """Return the bitfield attribute."""
        return self.bitfield

    def calc_bitfield_from_list(self, localities: list[int]) -> int:
        """Calculate the bitfield from a list of localities."""
        bitfield = 0b00000000

        # First, we need to validate all of the localities in the list.
        for value in localities:
            # If the value is in a bad range, we're done here.
            if not (0 <= value < 5) and not (32 <= value < 256):
                raise ValueError("Invalid locality '%d'!" % value)
            # An "upper" locality must be individual. Cannot combine with 0-4.
            if (32 <= value < 256) and len(localities) > 1:
                raise ValueError("Cannot combine locality '%d' with others!" % value)

        # If the list is empty... well, we're done.
        if len(localities) == 0:
            pass

        # Now, if we're an "upper" locality, that's a simple value.
        elif len(localities) == 1 and (32 <= localities[0] < 256):
            bitfield = localities[0]

        # We have to actually "think" to calculate the "lower" localities.
        else:
            for value in localities:
                bitfield |= 1 << value

        return bitfield

    def get_buffer_for_digest(self) -> str:
        r"""Serializes the primitive.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        # NOTE: We force big-endian to match the marshalling in the TPM.
        return struct.pack(">LB", t2d.TPM_CC_PolicyLocality, self.bitfield)


class PolicyCommandCode(object):
    """Policy Primitive to describe a Command code."""

    def __init__(self, command_code_string: Optional[str] = None) -> "PolicyCommandCode":
        """Init with the requested command code string."""
        # Check to make sure that a command_code can be found.
        str_command_code_string = str(command_code_string)
        command_code = t2d.CommandCode.get_code(str_command_code_string)
        if command_code is None:
            raise ValueError("Command code '%s' unknown!" % str_command_code_string)
        self.command_code_string = str_command_code_string

    def get_code(self) -> str:
        """Returns the command_code_string attribute."""
        return self.command_code_string

    def get_buffer_for_digest(self) -> str:
        r"""Serializes the primitive.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        # NOTE: We force big-endian to match the marshalling in the TPM.
        return struct.pack(
            ">LL",
            t2d.CommandCode.get_code("TPM_CC_PolicyCommandCode"),
            t2d.CommandCode.get_code(self.command_code_string),
        )



# @file package_table.py
# A module to associate the packages in a workspace with the repositories they come from.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A module to generate a table containing information about a package."""

from pathlib import Path
from typing import Any

import git

from edk2toollib.database import Package, Repository, Session
from edk2toollib.database.tables import TableGenerator
from edk2toollib.uefi.edk2.path_utilities import Edk2Path

GIT_EXTENSION = ".git"
DEC_EXTENSION = "*.dec"


class PackageTable(TableGenerator):
    """A Table Generator that associates packages with their repositories."""

    def __init__(self, *args: Any, **kwargs: Any) -> "PackageTable":
        """Initializes the Repository Table Parser.

        Args:
            args (any): non-keyword arguments
            kwargs (any): None

        """

    def get_repo_name(repo: git.Repo) -> str:
        """Get the name of the repository."""
        if "origin" in repo.remotes:
            return repo.remotes.origin.url.split("/")[-1].split(GIT_EXTENSION)[0].upper()
        elif len(repo.remotes) > 0:
            return repo.remotes[0].url.split("/")[-1].split(GIT_EXTENSION)[0].upper()
        return "BASE"

    def parse(self, session: Session, pathobj: Edk2Path, id: str, env: dict) -> None:
        """Glob for packages and insert them into the table."""
        try:
            repo = git.Repo(pathobj.WorkspacePath)
        except git.InvalidGitRepositoryError:
            return

        all_packages = {(pkg.name, pkg.path): pkg for pkg in session.query(Package).all()}
        all_repos = {(repo.name, repo.path): repo for repo in session.query(Repository).all()}

        packages_to_add = []
        for file in Path(pathobj.WorkspacePath).rglob(DEC_EXTENSION):
            pkg_name = file.parent.name
            containing_repo = PackageTable.get_repo_name(repo)
            repo_path = None

            for submodule in repo.submodules:
                if submodule.abspath in str(file):
                    containing_repo = submodule.name
                    repo_path = submodule.path
                    break

            repository = all_repos.setdefault(
                (containing_repo, repo_path), Repository(name=containing_repo, path=repo_path)
            )

            pkg_path = file.parent.relative_to(pathobj.WorkspacePath).as_posix()
            if (pkg_name, pkg_path) not in all_packages:
                package = all_packages.setdefault(
                    (pkg_name, pkg_path), Package(name=pkg_name, path=pkg_path, repository=repository)
                )
                packages_to_add.append(package)
        session.add_all(packages_to_add)



##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A collection of table generators that run against the workspace."""

from edk2toollib.database.edk2_db import TableGenerator  # noqa: F401
from edk2toollib.database.tables.environment_table import EnvironmentTable  # noqa: F401
from edk2toollib.database.tables.inf_table import InfTable  # noqa: F401
from edk2toollib.database.tables.instanced_fv_table import InstancedFvTable  # noqa: F401
from edk2toollib.database.tables.instanced_inf_table import InstancedInfTable  # noqa: F401
from edk2toollib.database.tables.package_table import PackageTable  # noqa: F401
from edk2toollib.database.tables.source_table import SourceTable  # noqa: F401
